from .base import Renderable
from .collection import RenderablesCollection
from .mesh import Mesh
from .pointcloud import PointCloud
from .primitives import MeshPrimitive, CubeMesh, CircleMesh, CylinderMesh, EllipsoidNURBS, SphereNURBS, CurveBezier



from .bsdf import PrincipledBSDFMaterial, GlossyBSDFMaterial, PrincipledBSDFWireframeMaterial
from .metal import MetalMaterial, MetalWireframeMaterial
from .plastic import PlasticMaterial, PlasticWireframeMaterial
from .emission import EmissionMaterial



import os
import shutil
import sys
import tempfile
import warnings
from contextlib import nullcontext
from pathlib import Path
from typing import Union, Sequence

import bpy
import numpy as np

os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2

from .cameras import PerspectiveCamera, OrthographicCamera
from .cameras.base import Camera
from .internal import Singleton
from .internal.io import catch_stdout
from .internal.types import Vector2d, Vector2di, Vector3d, Vector4d, RotationParams
from .internal import parser
from .lights import LightsCollection
from .renderables import RenderablesCollection
from memory_tempfile import MemoryTempfile
from typing import Optional


class Scene(metaclass=Singleton):
    def __init__(self):
        # Initialise Blender scene
        self.renderables = RenderablesCollection()
        self.lights = LightsCollection()
        self._camera = None
        self._reset_scene()
        self.tempfile = MemoryTempfile()
        self._postprocess_fn = lambda x: x

    @staticmethod
    def _set_default_blender_parameters():
        # Setup scene parameters
        scene = bpy.data.scenes[0]
        scene.use_nodes = True
        bpy.context.scene.world.use_nodes = False
        bpy.context.scene.render.engine = "CYCLES"
        bpy.context.scene.render.image_settings.color_mode = "RGBA"
        bpy.context.scene.render.image_settings.file_format = "PNG"
        bpy.context.scene.render.image_settings.quality = 100
        bpy.context.scene.render.image_settings.color_depth = "8"
        bpy.context.scene.world.color = (0, 0, 0)
        bpy.context.scene.render.film_transparent = True
        bpy.context.scene.cycles.filter_width = 0  # turn off anti-aliasing
        # Important if you want to get a pure color background (eg. white background)
        bpy.context.scene.view_settings.view_transform = "Standard"
        bpy.context.scene.cycles.samples = (
            128  # Default value, can be changed in .render
        )
        bpy.context.scene.frame_current = 0

    @staticmethod
    def _remove_all_objects():
        """Removes all objects from the scene. Previously used to remove the default cube"""
        bpy.ops.object.mode_set(mode="OBJECT")
        bpy.ops.object.select_all(action="SELECT")
        bpy.ops.object.delete(use_global=False)
        bpy.ops.outliner.orphans_purge()
        bpy.ops.outliner.orphans_purge()
        bpy.ops.outliner.orphans_purge()

    @staticmethod
    def _load_empty_scene():
        """Resets the scene to the empty state"""
        bpy.ops.wm.read_homefile(use_empty=True)
        bpy.ops.outliner.orphans_purge()

    def _reset_scene(self):
        """Resets the scene to the empty state"""
        with catch_stdout():
            self._load_empty_scene()
        scene = bpy.data.scenes[0]
        scene.world = bpy.data.worlds.new("BlendifyWorld")
        self._frame_number = 0
        self._set_default_blender_parameters()
        self.renderables._reset()
        self.lights._reset()
        self._camera = None

    def clear(self):
        """Clears the scene"""
        self._reset_scene()

    @property
    def camera(self) -> Camera:
        return self._camera

    def set_perspective_camera(
        self,
        resolution: Vector2di,
        focal_dist: float = None,
        fov_x: float = None,
        fov_y: float = None,
        center: Vector2d = None,
        near: float = 0.1,
        far: float = 100.0,
        tag: str = "camera",
        rotation_mode: str = "quaternionWXYZ",
        rotation: RotationParams = None,
        translation: Vector3d = (0, 0, 0),
        resolution_percentage: int = 100,
    ) -> PerspectiveCamera:
        """Set perspective camera in the scene. Replaces the previous scene camera, if it exists.
        One of focal_dist, fov_x or fov_y is required to set the camera parameters

        Args:
            resolution (Vector2di): (w, h), the resolution of the resulting image
            focal_dist (float, optional): Perspective Camera focal distance in millimeters (default: None)
            fov_x (float, optional): Camera lens horizontal field of view (default: None)
            fov_y (float, optional): Camera lens vertical field of view (default: None)
            center (Vector2d, optional): (x, y), horizontal and vertical shifts of the Camera (default: None)
            near (float, optional): Camera near clipping distance (default: 0.1)
            far (float, optional): Camera far clipping distance (default: 100)
            tag (str): name of the created object in Blender
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            resolution_percentage (int, optional):
        Returns:
            PerspectiveCamera: created camera
        """
        camera = PerspectiveCamera(
            resolution=resolution,
            focal_dist=focal_dist,
            fov_x=fov_x,
            fov_y=fov_y,
            center=center,
            near=near,
            far=far,
            tag=tag,
            rotation_mode=rotation_mode,
            rotation=rotation,
            translation=translation,
        )
        self._setup_camera(camera, resolution_percentage)
        return camera

    def set_opencv_camera(
        self,
        resolution: Vector2di,
        intrinsic_matrix: np.ndarray,
        extrinsic_matrix: np.ndarray,
        near: float = 0.1,
        far: float = 100.0,
        tag: str = "camera",
        resolution_percentage: int = 100,
    ):

        R = extrinsic_matrix[:3, :3]
        t = -R.T @ extrinsic_matrix[:3, 3]
        self.set_perspective_camera(
            resolution,
            rotation=R.T * [1, -1, -1],
            rotation_mode="rotmat",
            translation=t,
            focal_dist=intrinsic_matrix[0, 0],
            center=intrinsic_matrix[:2, 2] / resolution,
            near=near,
            far=far,
            tag=tag,
            resolution_percentage=resolution_percentage,
        )
        # self._postprocess_fn = distort_fn

    def set_orthographic_camera(
        self,
        resolution: Vector2di,
        ortho_scale: float = 1.0,
        near: float = 0.1,
        far: float = 100.0,
        tag: str = "camera",
        rotation_mode: str = "quaternionWXYZ",
        rotation: RotationParams = None,
        translation: Vector3d = (0, 0, 0),
        resolution_percentage: int = 100,
    ) -> OrthographicCamera:
        """Set orthographic camera in the scene. Replaces the previous scene camera, if it exists

        Args:
            resolution (Vector2di): (w, h), the resolution of the resulting image
            ortho_scale (float, optional): Orthographic Camera scale (similar to zoom) (default: 1.0)
            near (float, optional): Camera near clipping distance (default: 0.1)
            far (float, optional): Camera far clipping distance (default: 100)
            tag (str): name of the created object in Blender
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            resolution_percentage (int, optional):
        Returns:
            OrthographicCamera: created camera
        """
        camera = OrthographicCamera(
            resolution=resolution,
            ortho_scale=ortho_scale,
            far=far,
            near=near,
            tag=tag,
            rotation_mode=rotation_mode,
            rotation=rotation,
            translation=translation,
        )
        self._setup_camera(camera, resolution_percentage)
        return camera

    def _setup_camera(self, camera: Camera, resolution_percentage: int = 100):
        # Delete old camera
        if self._camera is not None:
            self._camera._blender_remove_object()
        # Set new camera
        self._camera = camera
        scene = bpy.data.scenes[0]
        scene.render.resolution_x = camera.resolution[0]
        scene.render.resolution_y = camera.resolution[1]
        scene.render.resolution_percentage = resolution_percentage

    @staticmethod
    def read_exr_distmap(path: str, dist_thresh: float = 1e4) -> np.ndarray:
        """Reads the distance map stored in EXR format, filters out all the values after a certain distance threshold.
        Requires OpenEXR to be installed in the system

        Args:
            path (str): path to the .exr file
            dist_thresh (float): distance clip threshold

        Returns:
            np.ndarray: distance map in numpy array format
        """
        data = cv2.imread(path, cv2.IMREAD_UNCHANGED)[:, :, 0]
        data[data > dist_thresh] = -np.inf
        return data

    @staticmethod
    def read_image(path: str) -> np.ndarray:
        """Reads the image stored in PNG or JPG format

        Args:
            path (str): path to the image file

        Returns:
            np.ndarray: image in numpy array format
        """
        return cv2.cvtColor(cv2.imread(path, cv2.IMREAD_UNCHANGED), cv2.COLOR_BGRA2RGBA)

    def render(
        self,
        filepath: Union[str, Path] = None,
        use_gpu: bool = True,
        samples: int = 128,
        save_depth: bool = False,
        save_albedo: bool = False,
        verbose: bool = False,
        use_denoiser: bool = False,
        aa_filter_width: float = 1.5,
        frame_number: int = None,
        adaptive_sampling_threshold: Optional[float] = None,
        color_dtype=np.uint8,
    ):
        """Start the Blender rendering process

        Automatically detects if shadow catcher objects are present in the scene by checking is_shadow_catcher property
        of Blender objects. For rendering with shadow catchers background is made white, because since Blender 3.0
        shadow catcher rendering pass is made to be multiplied with the background.

        Args:
            filepath (Union[str, Path]): path to the image (PNG) to render to, returns the image as numpy array if None
            use_gpu (bool): whether to render on GPU or not
            samples (bool): number of raytracing samples per pixel
            save_depth (bool): whether to save the depth in the separate file.
              If yes, the numpy array <filepath>.depth.npy will be created if filepath is set, otherwise appends the array to the output.
            save_albedo (bool): whether to save albedo (raw color information) in the separate file.
              If yes, the PNG image <filepath>.albedo.png with color information will be created
              if filepath is set, otherwise appends the array to the output.
            verbose (bool): whether to allow blender to log its status to stdout during rendering
            use_denoiser (bool): use openimage denoiser to denoise the result
            aa_filter_width (float): width of the anti-aliasing filter, set 0 to turn off
        """
        if self.camera is None:
            raise RuntimeError("Can't render without a camera")

        if frame_number is not None:
            self._frame_number = frame_number

        # setup anti-aliasing
        aa_filter_width = max(0, aa_filter_width)
        bpy.context.scene.cycles.filter_width = aa_filter_width
        bpy.context.scene.cycles.pixel_filter_type = "BLACKMAN_HARRIS"
        if save_depth and aa_filter_width != 0:
            warnings.warn(
                "Anti-aliasing filter is enabled. Saved depth will not be exact."
            )

        scene = bpy.data.scenes[0]
        scene.render.resolution_x = self.camera.resolution[0]
        scene.render.resolution_y = self.camera.resolution[1]
        scene.render.resolution_percentage = 100

        if color_dtype not in [np.uint8, np.uint16]:
            raise ValueError("Invalid rgb_dtype. Must be np.uint8 or np.uint16")
        bpy.context.scene.render.image_settings.color_depth = (
            "8" if color_dtype == np.uint8 else "16"
        )
        bpy.context.scene.camera = self.camera.blender_camera
        scene.cycles.sample_clamp_direct = 2.0

        # Setup denoising
        if use_denoiser:
            bpy.context.scene.cycles.use_denoising = True
            try:
                bpy.context.scene.cycles.denoiser = "OPTIX"
            except:
                bpy.context.scene.cycles.denoiser = "OPENIMAGEDENOISE"
            bpy.context.scene.view_layers[0].cycles.use_denoising = True

            # bpy.context.view_layer.cycles.denoising_store_passes = True
            bpy.context.scene.cycles.denoising_input_passes = (
                "RGB_ALBEDO_NORMAL"  # Options: 'RGB', 'RGB_ALBEDO', 'RGB_ALBEDO_NORMAL'
            )
            bpy.context.scene.cycles.denoising_prefilter = (
                "ACCURATE"  # Options: 'FAST', 'ACCURATE'
            )

        else:
            bpy.context.scene.cycles.use_denoising = False
            bpy.context.scene.view_layers[0].cycles.use_denoising = False
            bpy.context.view_layer.cycles.denoising_store_passes = False

        # Configure output
        bpy.context.scene.cycles.samples = samples
        if adaptive_sampling_threshold is not None:
            bpy.context.scene.cycles.use_adaptive_sampling = True
            bpy.context.scene.cycles.adaptive_threshold = adaptive_sampling_threshold
        else:
            bpy.context.scene.cycles.use_adaptive_sampling = False

        bpy.context.scene.view_layers["ViewLayer"].use_pass_combined = True
        bpy.context.scene.view_layers["ViewLayer"].use_pass_diffuse_color = True
        bpy.context.scene.view_layers["ViewLayer"].use_pass_z = True
        scene_node_tree = bpy.context.scene.node_tree

        for n in scene_node_tree.nodes:
            scene_node_tree.nodes.remove(n)
        render_layer = scene_node_tree.nodes.new(type="CompositorNodeRLayers")
        rendered_image = render_layer.outputs["Image"]

        # check if we have shadow catchers
        use_shadow_catcher = False
        for obj in bpy.data.objects:
            if obj.type != "LIGHT" and obj.is_shadow_catcher:
                use_shadow_catcher = True
                break

        # create output node
        if use_shadow_catcher:
            bpy.context.view_layer.cycles.use_pass_shadow_catcher = True
            alpha_over = scene_node_tree.nodes.new(type="CompositorNodeAlphaOver")
            scene_node_tree.links.new(
                render_layer.outputs["Shadow Catcher"], alpha_over.inputs[1]
            )
            scene_node_tree.links.new(rendered_image, alpha_over.inputs[2])

            output_image = scene_node_tree.nodes.new(type="CompositorNodeOutputFile")
            scene_node_tree.links.new(
                alpha_over.outputs["Image"], output_image.inputs["Image"]
            )
        else:
            bpy.context.view_layer.cycles.use_pass_shadow_catcher = False
            output_image = scene_node_tree.nodes.new(type="CompositorNodeOutputFile")
            scene_node_tree.links.new(rendered_image, output_image.inputs["Image"])

        if save_depth:
            output_depth = scene_node_tree.nodes.new(type="CompositorNodeOutputFile")
            output_depth.format.file_format = "OPEN_EXR"
            scene_node_tree.links.new(
                render_layer.outputs["Depth"], output_depth.inputs["Image"]
            )

        if save_albedo:
            output_albedo = scene_node_tree.nodes.new(type="CompositorNodeOutputFile")
            scene_node_tree.links.new(
                render_layer.outputs["DiffCol"], output_albedo.inputs["Image"]
            )

        if use_gpu:
            bpy.context.scene.cycles.device = "GPU"

            for scene in bpy.data.scenes:
                scene.cycles.device = "GPU"

            # Detect the appropriate GPU rendering mode
            rendering_mode_priority_list = ["OPTIX", "HIP", "ONEAPI", "CUDA"]
            rendering_preferences = bpy.context.preferences.addons["cycles"].preferences
            rendering_preferences.refresh_devices()
            devices = rendering_preferences.devices
            available_rendering_modes = set()
            for dev in devices:
                available_rendering_modes.add(dev.type)
            chosen_rendering_mode = "NONE"
            for mode in rendering_mode_priority_list:
                if mode in available_rendering_modes:
                    chosen_rendering_mode = mode
                    break

            # Set GPU rendering mode to detected one
            rendering_preferences.compute_device_type = chosen_rendering_mode

            # Optionally, list the devices before rendering
            # for dev in devices:
            #     print(f"ID:{dev.id} Name:{dev.name} Type:{dev.type} Use:{dev.use}")

        # Render
        bpy.context.scene.frame_current = self._frame_number
        render_to_ram = filepath is None
        with (
            self.tempfile.TemporaryDirectory() if render_to_ram else nullcontext()
        ) as tmpdir:
            if render_to_ram:
                basepath = tmpdir
                filename = "result.png"
                filepath = Path(tmpdir) / filename
            else:
                filepath = Path(filepath)
                basepath = str(filepath.parent.absolute())
                filename = filepath.stem

            scene.render.filepath = str(basepath)

            tempfile.mktemp()
            temp_filesuffix = next(tempfile._get_candidate_names())
            temp_filepath = str(filepath) + "." + temp_filesuffix
            render_suffixes = [f".color.######.png"]
            if save_depth:
                render_suffixes.append(f".depth.######.exr")
            if save_albedo:
                render_suffixes.append(f".albedo.######.png")
            while self.check_any_exists(temp_filepath, render_suffixes):
                temp_filesuffix = next(tempfile._get_candidate_names())
                temp_filepath = str(filepath) + "." + temp_filesuffix
            temp_filename = os.path.basename(temp_filepath)
            output_image.base_path = basepath
            output_image.file_slots[0].path = temp_filename + ".color."
            if save_depth:
                output_depth.file_slots[0].path = temp_filename + ".depth."
            if save_albedo:
                output_albedo.file_slots[0].path = temp_filename + ".albedo."

            with catch_stdout(skip=verbose):
                bpy.ops.render.render(write_still=False)

            if render_to_ram:
                image_data = self.read_image(
                    temp_filepath + f".color.{self._frame_number:04d}.png"
                )
                outputs = [image_data]
                if save_depth:
                    distmap = self.read_exr_distmap(
                        temp_filepath + f".depth.{self._frame_number:04d}.exr",
                        dist_thresh=self.camera.far * 1.1,
                    )
                    depthmap = self.camera.distance2depth(distmap)
                    outputs.append(depthmap)
                if save_albedo:
                    albedomap = self.read_image(
                        temp_filepath + f".albedo.{self._frame_number:04d}.png"
                    )
                    outputs.append(albedomap)
            else:
                shutil.move(
                    temp_filepath + f".color.{self._frame_number:04d}.png", filepath
                )
                output_image.file_slots[0].path = filename
                if save_depth:
                    distmap = self.read_exr_distmap(
                        temp_filepath + f".depth.{self._frame_number:04d}.exr",
                        dist_thresh=self.camera.far * 1.1,
                    )
                    depthmap = self.camera.distance2depth(distmap)
                    np.save(os.path.splitext(filepath)[0] + ".depth.npy", depthmap)
                    os.remove(temp_filepath + f".depth.{self._frame_number:04d}.exr")
                    output_depth.file_slots[0].path = filename + ".depth."
                if save_albedo:
                    shutil.move(
                        temp_filepath + f".albedo.{self._frame_number:04d}.png",
                        os.path.splitext(filepath)[0] + ".albedo.png",
                    )
                    output_albedo.file_slots[0].path = filename + ".albedo."
                outputs = None

            self._frame_number += 1
            if outputs is not None and len(outputs) == 1:
                return outputs[0]
            else:
                return outputs

    def preview(
        self,
        filepath: Union[str, Path] = None,
        save_depth: bool = False,
        save_albedo: bool = False,
        verbose: bool = False,
        fast: bool = False,
        frame_number: Optional[int] = None,
    ):
        """Renders a scene using Blender's OpenGL renderer. Linux and MacOS Only.

        Args:
            filepath (Union[str, Path]): path to the image (PNG) to render to, returns the image as numpy array if None
            use_gpu (bool): whether to render on GPU or not
            samples (bool): number of raytracing samples per pixel
            save_depth (bool): whether to save the depth in the separate file.
              If yes, the numpy array <filepath>.depth.npy will be created if filepath is set, otherwise appends the array to the output.
            save_albedo (bool): whether to save albedo (raw color information) in the separate file.
              If yes, the PNG image <filepath>.albedo.png with color information will be created
              if filepath is set, otherwise appends the array to the output.
            verbose (bool): whether to allow blender to log its status to stdout during rendering
            use_denoiser (bool): use openimage denoiser to denoise the result
            fast (bool): whether to use fast and colorless preview mode (workbench engine)
        """
        assert sys.platform.startswith("linux") or sys.platform.startswith(
            "darwin"
        ), "Preview is only supported on Linux and MacOS"

        # Switch to OpenGL renderer
        if not fast:
            bpy.context.scene.render.engine = "BLENDER_EEVEE"
        else:
            bpy.context.scene.render.engine = "BLENDER_WORKBENCH"

        if self.camera is None:
            raise RuntimeError("Can't render without a camera")

        if frame_number is not None:
            self._frame_number = frame_number
        else:
            self._frame_number += 1

        scene = bpy.data.scenes[0]
        scene.render.resolution_x = self.camera.resolution[0]
        scene.render.resolution_y = self.camera.resolution[1]
        scene.render.resolution_percentage = 100

        bpy.context.scene.camera = self.camera.blender_camera
        # bpy.context.object.data.dof.focus_object = object

        bpy.context.scene.view_layers["ViewLayer"].use_pass_combined = True
        bpy.context.scene.view_layers["ViewLayer"].use_pass_diffuse_color = True
        bpy.context.scene.view_layers["ViewLayer"].use_pass_z = True
        scene_node_tree = bpy.context.scene.node_tree

        for n in scene_node_tree.nodes:
            scene_node_tree.nodes.remove(n)
        render_layer = scene_node_tree.nodes.new(type="CompositorNodeRLayers")
        output_image = scene_node_tree.nodes.new(type="CompositorNodeOutputFile")
        scene_node_tree.links.new(
            render_layer.outputs["Image"], output_image.inputs["Image"]
        )

        if save_depth:
            output_depth = scene_node_tree.nodes.new(type="CompositorNodeOutputFile")
            output_depth.format.file_format = "OPEN_EXR"
            scene_node_tree.links.new(
                render_layer.outputs["Depth"], output_depth.inputs["Image"]
            )

        if save_albedo:
            output_albedo = scene_node_tree.nodes.new(type="CompositorNodeOutputFile")
            scene_node_tree.links.new(
                render_layer.outputs["DiffCol"], output_albedo.inputs["Image"]
            )

        # Render
        bpy.context.scene.frame_current = self._frame_number

        render_to_ram = filepath is None
        with (
            self.tempfile.TemporaryDirectory() if render_to_ram else nullcontext()
        ) as tmpdir:
            if render_to_ram:
                basepath = tmpdir
                filename = "result.png"
                filepath = Path(tmpdir) / filename
            else:
                filepath = Path(filepath)
                basepath = str(filepath.parent.absolute())
                filename = filepath.stem

            scene.render.filepath = str(filepath.parent)
            temp_filesuffix = next(tempfile._get_candidate_names())
            temp_filepath = str(filepath) + "." + temp_filesuffix
            render_suffixes = [f".color.{self._frame_number:04d}.png"]
            if save_depth:
                render_suffixes.append(f".depth.{self._frame_number:04d}.exr")
            if save_albedo:
                render_suffixes.append(f".albedo.{self._frame_number:04d}.png")
            while self.check_any_exists(temp_filepath, render_suffixes):
                temp_filesuffix = next(tempfile._get_candidate_names())
                temp_filepath = str(filepath) + "." + temp_filesuffix
            temp_filename = os.path.basename(temp_filepath)
            output_image.base_path = basepath
            output_image.file_slots[0].path = temp_filename + ".color."
            if save_depth:
                output_depth.file_slots[0].path = temp_filename + ".depth."
            if save_albedo:
                output_albedo.file_slots[0].path = temp_filename + ".albedo."

            with catch_stdout(skip=verbose):
                bpy.ops.render.render(write_still=False)

            if render_to_ram:
                image_data = self.read_image(
                    temp_filepath + f".color.{self._frame_number:04d}.png"
                )
                outputs = [image_data]
                if save_depth:
                    distmap = self.read_exr_distmap(
                        temp_filepath + f".depth.{self._frame_number:04d}.exr",
                        dist_thresh=self.camera.far * 1.1,
                    )
                    depthmap = self.camera.distance2depth(distmap)
                    outputs.append(depthmap)
                if save_albedo:
                    albedomap = self.read_image(
                        temp_filepath + f".albedo.{self._frame_number:04d}.png"
                    )
                    outputs.append(albedomap)
                if len(outputs) == 1:
                    return outputs[0]
                else:
                    return outputs
            else:
                shutil.move(
                    temp_filepath + f".color.{self._frame_number:04d}.png", filepath
                )
                output_image.file_slots[0].path = filename
                if save_depth:
                    distmap = self.read_exr_distmap(
                        temp_filepath + f".depth.{self._frame_number:04d}.exr",
                        dist_thresh=self.camera.far * 1.1,
                    )
                    depthmap = self.camera.distance2depth(distmap)
                    np.save(os.path.splitext(filepath)[0] + ".depth.npy", depthmap)
                    os.remove(temp_filepath + f".depth.{self._frame_number:04d}.exr")
                    output_depth.file_slots[0].path = filename + ".depth."
                if save_albedo:
                    shutil.move(
                        temp_filepath + f".albedo.{self._frame_number:04d}.png",
                        os.path.splitext(filepath)[0] + ".albedo.png",
                    )
                    output_albedo.file_slots[0].path = filename + ".albedo."

        # Return to Cycles renderer
        bpy.context.scene.render.engine = "CYCLES"

    @staticmethod
    def check_any_exists(fileprefix: str, filesuffixes: Sequence[str]) -> bool:
        """Check if any of the combinations of <fileprefix>+<any filesuffix> exist in the filesystem

        Args:
            fileprefix (str): single file prefix, can the full path or local name
            filesuffixes (Sequence[str]): a sequence of file suffixes to choose from

        Returns:
            bool: True is any of the combinations exists in the filesystem, False otherwise
        """
        for filesuffix in filesuffixes:
            fullpath = fileprefix + filesuffix
            if os.path.exists(fullpath):
                return True
        return False

    @staticmethod
    def export(
        path: Union[str, Path],
        include_file_textures: bool = True,
        verbose: bool = False,
    ):
        """Export the current scene to the .blend file

        Args:
            path (Union[str, Path]): path to the target .blend file
            include_file_textures (bool): whether to write textures loaded from external files inside .blend file
            verbose (bool): whether to allow blender to log its status to stdout during exporting
        """
        # hack to overcome Blender error message "BKE_bpath_relative_convert: basedir='', this is a bug"
        path = str(os.path.abspath(path))

        # Create folder
        Path(path).parent.mkdir(parents=True, exist_ok=True)

        with catch_stdout(skip=verbose):
            if include_file_textures:
                bpy.ops.file.pack_all()
            bpy.ops.wm.save_as_mainfile(filepath=path)

    def attach_blend(self, path: Union[str, Path], with_camera: bool = False):
        """Append objects and materials from the existing .blend file to the scene.
        The only two modalities that can be added to blendify Scene are lights and optionally camera,
        others (renderable objects, materials, etc.) are only appended.
        The appended modalities will only be present in the internal Blender structures,
        but will not be added to the blendify Scene class structure.
        However, they will appear on rendering and in the exported .blend files

        Args:
            path: path to the .blend file to append the contents from
            with_camera: parse camera parameters from the .blend file or keep existing camera
        """
        main_scene_name = bpy.data.scenes.keys()[0]
        materials = []
        with bpy.data.libraries.load(str(path), link=False) as (data_from, data_to):
            for name in data_from.materials:
                materials.append({"name": name})

            if with_camera:
                # Check number of cameras by looking at cameras settings
                assert (
                    len(data_from.cameras) == 1
                ), f"Expect to have only single camera in .blend, got {len(data_from.cameras)}"

                # Parse resolution parameters from scene
                assert (
                    len(data_from.scenes) == 1
                ), f"Expect to have only single scene in .blend, got {len(data_from.scenes)}"

            data_to.scenes = data_from.scenes

        if with_camera:
            # Parse resolution
            res_x = data_to.scenes[0].render.resolution_x
            res_y = data_to.scenes[0].render.resolution_y
            resolution_percentage = data_to.scenes[0].render.resolution_percentage
            resolution = np.array([res_x, res_y])

            # Delete old camera
            if self._camera is not None:
                self._camera._blender_remove_object()
        else:
            # Remove camera from the imported scene
            camera_obj = data_to.scenes[0].camera

            with bpy.context.temp_override(selected_objects=[camera_obj]):
                bpy.ops.object.delete()

        # Add materials to the current scene
        if len(materials) > 0:
            bpy.ops.wm.append(
                directory=str(path) + "/Material/", files=materials, link=True
            )

        # Recursively copy collection
        main_scene = bpy.data.scenes[main_scene_name]
        import_scene = data_to.scenes[0]
        parser.move_collection(main_scene.collection, import_scene.collection)

        # Remove scene
        bpy.data.scenes.remove(import_scene, do_unlink=True)

        # Parse selected objects (lights and camera)
        bpy.context.view_layer.update()
        for obj in bpy.data.objects:
            if with_camera and obj.type == "CAMERA":
                camera_type, camera_dict = parser.parse_camera_from_blendfile(
                    obj, resolution
                )

                # Remove current camera, because we need to recreate it
                bpy.data.cameras.remove(bpy.data.cameras[0])

                if camera_type == "ORTHO":
                    self.set_orthographic_camera(
                        resolution_percentage=resolution_percentage, **camera_dict
                    )
                elif camera_type == "PERSP":
                    self.set_perspective_camera(
                        resolution_percentage=resolution_percentage, **camera_dict
                    )
                else:
                    raise NotImplementedError(f"Unsupported camera type {camera_type}")
            elif obj.type == "LIGHT":
                light_type, light_dict = parser.parse_light_from_blendfile(obj)

                # Remove light before re-creaing it
                with bpy.context.temp_override(selected_objects=[obj]):
                    bpy.ops.object.delete()

                if light_type == "POINT":
                    self.lights.add_point(**light_dict)
                elif light_type == "SUN":
                    self.lights.add_sun(**light_dict)
                elif light_type == "SPOT":
                    self.lights.add_spot(**light_dict)
                elif light_type == "AREA":
                    self.lights.add_area(**light_dict)
                else:
                    raise NotImplementedError(f"Unsupported light type {light_type}")



from typing import Union

import bpy

from ..renderables.base import Renderable


def set_shadow_catcher(obj: Union[str, Renderable], state=False):
    """Set object is_shadow_catcher and various visibility_* properties to make
    object act as a shadow catcher or revert the changes.

    Args:
        obj (Union[str, Renderable]): tag of the Blender object or instance of Renderable
        state (bool): if True make obj a shadow catcher, otherwise make it a regular object
    """
    if not isinstance(obj, str):
        obj = obj.tag

    # select object
    bpy.data.objects[obj].select_set(True)
    bpy.context.view_layer.objects.active = bpy.data.objects[obj]
    shadow_catcher = bpy.context.active_object

    # set / unset shadow catcher properties
    shadow_catcher.is_shadow_catcher = state
    shadow_catcher.cycles.is_shadow_catcher = state
    shadow_catcher.visible_glossy = not state
    shadow_catcher.visible_diffuse = not state
    shadow_catcher.visible_transmission = not state
    shadow_catcher.visible_volume_scatter = not state

    # deselect object
    bpy.data.objects[obj].select_set(False)



import numpy as np
from typing import Union

from .base import ColorsMetadata, Colors
from ..internal.types import Vector3d, Vector4d


class VertexColors(Colors):
    """A container which stores a color information for each vertex of an object
    (vertex colors are interpolated over the faces)
    """

    def __init__(self, vertex_colors: np.ndarray):
        """Create the vertex color container

        Args:
            vertex_colors (np.ndarray): numpy array of size (N,3) for RGB or (N,4) for RGBD colors
        """
        super().__init__()
        assert (np.ndim(vertex_colors) == 2 and 3 <= vertex_colors.shape[1] <= 4), \
            f"Expected colors array of shape (N,3) or (N,4) got shape {vertex_colors.shape}"
        assert vertex_colors.dtype in [np.float32, np.float64], \
            "Colors should be stored as floating point numbers (np.float32 or np.float64)"
        assert np.all(vertex_colors >= 0) and np.all(vertex_colors <= 1), "Colors should be in range [0.0, 1.0]"
        has_alpha = vertex_colors.shape[1] == 4
        if not has_alpha:
            vertex_colors = np.hstack([vertex_colors, np.ones((vertex_colors.shape[0], 1), dtype=vertex_colors.dtype)])
        self._vertex_colors = vertex_colors.copy()
        self._metadata = ColorsMetadata(
            type=self.__class__,
            has_alpha=has_alpha,
            color=None,
            texture=None
        )

    @property
    def vertex_colors(self) -> np.ndarray:
        """Get current colors

        Returns:
            np.ndarray: current vertex colors
        """
        return self._vertex_colors


class UniformColors(Colors):
    """A container which stores a single uniform color for the whole object
    """

    def __init__(self, uniform_color: Union[Vector3d, Vector4d]):
        """Create the uniform color container

        Args:
            uniform_color (Vector3d): a color in RGB format (to change alpha, use 'alpha' material property instead)
        """
        super().__init__()
        uniform_color = np.array(uniform_color)
        assert len(uniform_color) in (3, 4), "Color should be in RGB or RGBA format"
        assert uniform_color.max() <= 1. and uniform_color.min() >= 0., "Color values should be in [0,1] range"
        self._color = uniform_color
        self._metadata = ColorsMetadata(
            type=self.__class__,
            has_alpha=self._color.shape[0] == 4,
            color=self._color,
            texture=None
        )

    @property
    def color(self) -> np.ndarray:
        """Get current color

        Returns:
            np.ndarray: current color
        """
        return self._color



from typing import Union, Tuple

import bpy
import numpy as np

BlenderGroup = Union[bpy.types.Collection, bpy.types.Object]
Vector2d = Union[np.ndarray, Tuple[float, float]]
Vector2di = Union[np.ndarray, Tuple[int, int]]
Vector3d = Union[np.ndarray, Tuple[float, float, float]]
Vector3di = Union[np.ndarray, Tuple[int, int, int]]
Vector4d = Union[np.ndarray, Tuple[float, float, float, float]]
Mat3x3 = Union[np.ndarray, Tuple[Tuple[float, float, float], Tuple[float, float, float], Tuple[float, float, float]]]
RotationParams = Union[Vector4d, Vector3d, Mat3x3, "Positionable", Tuple[Vector3d, float], Tuple["Positionable", float]]


import os
import sys
from io import open, UnsupportedOperation


class catch_stdout:
    """Stdout hook to ignore or reroute the blender output
    """

    def __init__(self, skip=False, logfile=None):
        """Creates a stdout hook object
        Args:
            skip: whether to skip the hook and pass the output to stdout
            logfile: filename to pass the output to (None to ignore the output completely)
        """
        self.skip = skip or not self._check_if_hook_possible()
        self.logfile_path = logfile if logfile is not None else os.devnull

    def _check_if_hook_possible(self):
        try:
            sys.stdout.fileno()
        except UnsupportedOperation:
            return False
        return True

    def set_hook(self):
        if self.skip:
            return
        self.logfile = open(self.logfile_path, 'w')
        self.logfile_fd = self.logfile.fileno()
        self.stdout_fd = sys.stdout.fileno()
        self.saved_stdout_fd = os.dup(self.stdout_fd)
        sys.stdout.flush()
        os.close(self.stdout_fd)
        os.dup2(self.logfile_fd, self.stdout_fd)
        self.capture_internal_stdout = False
        if sys.stdout.fileno() != sys.__stdout__.fileno():
            self.capture_internal_stdout = True
            self.internal_stdout_fd = sys.__stdout__.fileno()
            self.saved_internal_stdout_fd = os.dup(self.internal_stdout_fd)
            sys.__stdout__.flush()
            os.close(self.internal_stdout_fd)
            os.dup2(self.logfile_fd, self.internal_stdout_fd)

    def release_hook(self):
        if self.skip:
            return
        os.close(self.stdout_fd)
        os.dup2(self.saved_stdout_fd, self.stdout_fd)
        os.close(self.saved_stdout_fd)
        if self.capture_internal_stdout:
            os.close(self.internal_stdout_fd)
            os.dup2(self.saved_internal_stdout_fd, self.internal_stdout_fd)
            os.close(self.saved_internal_stdout_fd)
        self.logfile.close()

    def __enter__(self):
        self.set_hook()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release_hook()



from .base import Light
from ..internal.types import Vector3d


class PointLight(Light):
    def __init__(
        self,
        strength: float,
        shadow_soft_size: float,
        color: Vector3d,
        tag: str,
        cast_shadows: bool = True,
        **kwargs
    ):
        """Creates PointLight light source in Blender

        Args:
            strength (float): strength of the light source emitted over the entire area of the light in all directions
            shadow_soft_size (float): light size for ray shadow sampling (Raytraced shadows) in [0, inf]
            color (Vector3d): color of the light source
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        blender_light = self._blender_create_light(tag, "POINT")
        super().__init__(**kwargs, tag=tag, blender_object=blender_light)
        self.color = color
        self.strength = strength
        self.cast_shadows = cast_shadows
        self.shadow_soft_size = shadow_soft_size

    @property
    def shadow_soft_size(self) -> float:
        return self.blender_light.data.shadow_soft_size

    @shadow_soft_size.setter
    def shadow_soft_size(self, val: float):
        self.blender_light.data.shadow_soft_size = val


class DirectionalLight(Light):
    def __init__(
        self,
        strength: float,
        angular_diameter: float,
        color: Vector3d,
        tag: str,
        cast_shadows: bool = True,
        **kwargs
    ):
        """Creates DirectionalLight light source in Blender

        Args:
            strength (float): strength of the light source in watts per meter squared (W/m^2)
            angular_diameter (float): angular diameter of the Sun as seen from the Earth in [0, 3.14159]
            color (Vector3d): color of the light source
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        blender_light = self._blender_create_light(tag, "SUN")
        super().__init__(**kwargs, tag=tag, blender_object=blender_light)
        self.color = color
        self.strength = strength
        self.cast_shadows = cast_shadows
        self.angular_diameter = angular_diameter

    @property
    def angular_diameter(self) -> float:
        return self.blender_light.data.angle

    @angular_diameter.setter
    def angular_diameter(self, val: float):
        self.blender_light.data.angle = val


class SpotLight(Light):
    def __init__(
        self,
        strength: float,
        spot_size: float,
        spot_blend: float,
        color: Vector3d,
        shadow_soft_size: float,
        tag: str,
        cast_shadows: bool = True,
        **kwargs
    ):
        """Creates SpotLight light source in Blender

        Args:
            strength (float): strength of the light source that light would emit over its entire area if
                it wasn't limited by the spot angle
            spot_size (float): angle of the spotlight beam in [0.0174533, 3.14159]
            spot_blend (float): the softness of the spotlight edge in [0, 1]
            color (Vector3d): color of the light source
            shadow_soft_size (float): light size for ray shadow sampling (Raytraced shadows) in [0, inf]
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        # spot_size: Angle of the spotlight beam (float in [0.0174533, 3.14159]) default 0.785398
        # spot_blend: The softness of the spotlight edge (float in [0, 1]) default 0.15
        blender_light = self._blender_create_light(tag, "SPOT")
        super().__init__(**kwargs, tag=tag, blender_object=blender_light)
        self.color = color
        self.strength = strength
        self.spot_size = spot_size
        self.spot_blend = spot_blend
        self.cast_shadows = cast_shadows
        self.shadow_soft_size = shadow_soft_size

    @property
    def spot_size(self) -> float:
        return self.blender_light.data.spot_size

    @spot_size.setter
    def spot_size(self, val: float):
        self.blender_light.data.spot_size = val

    @property
    def spot_blend(self) -> float:
        return self.blender_light.data.spot_blend

    @spot_blend.setter
    def spot_blend(self, val: float):
        self.blender_light.data.spot_blend = val

    @property
    def shadow_soft_size(self) -> float:
        return self.blender_light.data.shadow_soft_size

    @shadow_soft_size.setter
    def shadow_soft_size(self, val: float):
        self.blender_light.data.shadow_soft_size = val



from abc import ABC, abstractmethod
from typing import NamedTuple, Optional, Union, Sequence

import bpy

from ..internal.types import Vector3d


class ColorsMetadata(NamedTuple):
    type: type
    color: Optional[Vector3d]
    has_alpha: bool
    texture: Optional[bpy.types.Image]
    interpolation: Optional[str] = None
    frame_start: Optional[int] = None

    def __del__(self):
        if self.texture is not None:
            if not self.texture.users:
                bpy.data.images.remove(self.texture)


class Colors(ABC):
    """An abstract container template for storing the object coloring information
    """

    @abstractmethod
    def __init__(self):
        self._metadata: Optional[ColorsMetadata] = None

    @property
    def metadata(self) -> ColorsMetadata:
        return self._metadata


ColorsList = Sequence[Colors]



from typing import Dict, Iterable, Union

import bpy
import numpy as np

from .base import Light
from .common import PointLight, DirectionalLight, SpotLight
from .area import AreaLight, SquareAreaLight, CircleAreaLight, RectangleAreaLight, EllipseAreaLight
from ..internal import Singleton
from ..internal.types import Vector2d, Vector3d, RotationParams


class LightsCollection(metaclass=Singleton):
    def __init__(self):
        self._lights: Dict[str, Light] = dict()
        self._background_light_nodes = None

    def remove_background_light(self):
        """
        Remove background light from the scene. Also removes ShaderNodeBackground if
        it was added via scene.attach_blend.

        Returns:
            None
        """
        world = bpy.context.scene.world
        world.use_nodes = True
        world_tree = world.node_tree

        if self._background_light_nodes is not None:
            for node_name in self._background_light_nodes:
                node = world_tree.nodes.get(node_name)
                world_tree.nodes.remove(node)
            self._background_light_nodes = None
        else:
            for node in world_tree.nodes:
                if node.type == 'BACKGROUND':
                    world_tree.nodes.remove(node)

    def set_background_light(self, strength: float = 1.0, color: Vector3d = (1.0, 1.0, 1.0)):
        """
        Set background light in a scene. Can be used to create ambient lightning in a scene.

        Returns:
            None
        """
        color = np.array(color)
        assert color.dtype in [np.float32, np.float64], \
            "Color should be stored as floating point numbers (np.float32 or np.float64)"
        world = bpy.context.scene.world
        world.use_nodes = True
        world_tree = world.node_tree

        self.remove_background_light()
        # Create light node and output node
        bg_node = world_tree.nodes.new(type='ShaderNodeBackground')
        for node in world_tree.nodes:
            if node.type == 'OUTPUT_WORLD':
                output_node = node
                break
        else:
            output_node = world_tree.nodes.new(type='ShaderNodeOutputWorld')

        # Set light color
        if len(color == 3):
            color = (*color, 1.0)
        bg_node.inputs['Color'].default_value = color

        # Link light to output and set strength
        world_tree.links.new(bg_node.outputs['Background'], output_node.inputs['Surface'])
        bg_node.inputs['Strength'].default_value = strength

        # Save created nodes to remove them later
        self._background_light_nodes = [bg_node.name, output_node.name]

    def add_point(
            self,
            strength: float = 100,
            shadow_soft_size: float = 0.25,
            color: Vector3d = (1.0, 1.0, 1.0),
            cast_shadows=True,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag=None
    ) -> PointLight:
        """Add PointLight light source to the scene

        Args:
            strength (float, optional): strength of the light source emitted over the entire area of the light
                in all directions (default: 100)
            shadow_soft_size (float, optional): light size for ray shadow sampling (Raytraced shadows)
                in [0, inf] (default: 0.25)
            color (Vector3d, optional): color of the light source (default: (1.0, 1.0, 1.0))
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created collection in Blender to represent the point cloud.
                If None is passed the tag is automatically generated (default: None)

        Returns:
            PointLight: created and added to the scene object
        """
        tag = self._process_tag(tag, "Point")
        light = PointLight(
            color=color, strength=strength, shadow_soft_size=shadow_soft_size, cast_shadows=cast_shadows,
            rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
        )
        self._lights[tag] = light
        return light

    def add_sun(
            self,
            strength: float = 10,
            angular_diameter: float = 0.00918043,
            color: Vector3d = (1.0, 1.0, 1.0),
            cast_shadows=True,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag=None
    ) -> DirectionalLight:
        """Add DirectionalLight light source to the scene

        Args:
            strength (float, optional): strength of the light source in watts per meter squared (W/m^2) (default: 100)
            angular_diameter (float, optional): angular diameter of the Sun as seen from the Earth
                in [0, 3.14159] (default: 0.00918043)
            color (Vector3d, optional): color of the light source (default: (1.0, 1.0, 1.0))
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created collection in Blender to represent the point cloud.
                If None is passed the tag is automatically generated (default: None)

        Returns:
            DirectionalLight: created and added to the scene object
        """
        # angular_diameter: Angular diameter of the Sun as seen from the Earth,  [0, 3.14159]
        tag = self._process_tag(tag, "Sun")
        light = DirectionalLight(
            color=color, strength=strength, angular_diameter=angular_diameter, cast_shadows=cast_shadows,
            rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
        )
        self._lights[tag] = light
        return light

    def add_spot(
            self,
            strength: float = 100,
            spot_size: float = 0.785398,
            spot_blend: float = 0.15,
            shadow_soft_size: float = 0.25,
            color: Vector3d = (1.0, 1.0, 1.0),
            cast_shadows=True,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag=None
    ) -> SpotLight:
        """Add SpotLight light source to the scene

        Args:
            strength (float, optional): strength of the light source that light would emit over its entire area if
                it wasn't limited by the spot angle (default: 100)
            spot_size (float, optional): angle of the spotlight beam in [0.0174533, 3.14159] (default: 0.785398)
            spot_blend (float, optional): the softness of the spotlight edge in [0, 1] (default: 0.15)
            shadow_soft_size (float, optional): light size for ray shadow sampling (Raytraced shadows)
                in [0, inf] (default: 0.25)
            color (Vector3d, optional): color of the light source (default: (1.0, 1.0, 1.0))
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created collection in Blender to represent the point cloud.
                If None is passed the tag is automatically generated (default: None)

        Returns:
            SpotLight: created and added to the scene object
        """
        tag = self._process_tag(tag, "Spot")
        light = SpotLight(
            color=color, strength=strength, spot_size=spot_size, spot_blend=spot_blend, cast_shadows=cast_shadows,
            shadow_soft_size=shadow_soft_size, rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
        )
        self._lights[tag] = light
        return light

    def add_area(
            self,
            shape: str,
            size: Union[float, Vector2d],
            strength: float = 100,
            color: Vector3d = (1.0, 1.0, 1.0),
            cast_shadows=True,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> AreaLight:
        """Add AreaLight light source to the scene. Shape of the area is controlled by shape parameter

        Args:
            shape (str): type of the AreaLight, one of: [square, circle, rectangle, ellipse]
            size (Union[float, Vector2d]): size of the area of the area light for circle and square or
                [x, y] sizes of the area light for rectangle and ellipse
            strength (float, optional): strength of the light source emitted over the entire area of the light
                in all directions (default: 100)
            color (Vector3d, optional): color of the light source (default: (1.0, 1.0, 1.0))
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created collection in Blender to represent the point cloud.
                If None is passed the tag is automatically generated (default: None)

        Returns:
            AreaLight: subclass of AreaLight (depending on shape), created and added to the scene
        """
        tag = self._process_tag(tag, "Area")
        if shape == "square":
            light = SquareAreaLight(
                size=size, color=color, strength=strength, cast_shadows=cast_shadows,
                rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
            )
        elif shape == "circle":
            light = CircleAreaLight(
                size=size, color=color, strength=strength, cast_shadows=cast_shadows,
                rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
            )
        elif shape == "rectangle":
            light = RectangleAreaLight(
                size=size, color=color, strength=strength, cast_shadows=cast_shadows,
                rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
            )
        elif shape == "ellipse":
            light = EllipseAreaLight(
                size=size, color=color, strength=strength, cast_shadows=cast_shadows,
                rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
            )
        else:
            raise NotImplementedError(f"Unknown AreaLight shape: {shape}")
        self._lights[tag] = light
        return light

    def _process_tag(self, tag: str, default_prefix: str = "Light"):
        lights_keys = self._lights.keys()

        if tag is None:
            _tag = default_prefix + "_{:03d}"
            index = 0
            while _tag.format(index) in lights_keys:
                index += 1
            tag = _tag.format(index)
        elif tag in lights_keys:
            raise RuntimeError(f"Object with tag {tag} is already in collection.")

        return tag

    def keys(self):
        return self._lights.keys()

    def values(self):
        return self._lights.values()

    def items(self):
        return self._lights.items()

    def remove(self, obj_or_tag: Union[Light, str]):
        assert isinstance(obj_or_tag, (Light, str)), "Only Light object or it's tag is allowed"
        if isinstance(obj_or_tag, str):
            tag = obj_or_tag
        else:
            obj = obj_or_tag
            tag = obj.tag
        self.__delitem__(tag)

    def __getitem__(self, key: str) -> Light:
        return self._lights[key]

    def __delitem__(self, key: str):
        self._lights[key]._blender_remove_object()
        del self._lights[key]

    def __iter__(self) -> Iterable:
        return iter(self._lights)

    def __len__(self) -> int:
        return len(self._lights)

    def _reset(self):
        self._lights = dict()
        self.remove_background_light()

    # def __str__(self):
    #     return str(self.__dict__)
    #
    # def __repr__(self):
    #     return '{}, D({})'.format(super(D, self).__repr__(), self.__dict__)


import numpy as np
import smplx
import torch


class SMPLWrapper:
    """A wrapper for the SMPL model
    """

    def __init__(self, smpl_root: str, gender: str, shape_params: np.ndarray, device: torch.device = None):
        self.device = torch.device(device if device is not None else "cpu")
        self.smpl_root = smpl_root
        self.shape_params = self._preprocess_param(shape_params)
        self.smpl = smplx.create(self.smpl_root, model_type = 'smpl', gender=gender, betas=self.shape_params).to(self.device)
        self.faces = np.array(self.smpl.faces, dtype=np.int64)

    @staticmethod
    def _center_output(smpl_model, params, smpl_output):
        """Center the SMPL model local coordinate system around the root joint"""
        if 'transl' in params and params['transl'] is not None:
            transl = params['transl']
        else:
            transl = None
        apply_trans = transl is not None or hasattr(smpl_model, 'transl')
        if transl is None and hasattr(smpl_model, 'transl'):
            transl = smpl_model.transl
        diff = -smpl_output.joints[:, 0, :]
        if apply_trans:
            diff = diff + transl
        smpl_output.joints = smpl_output.joints + diff.view(-1, 1, 3)
        smpl_output.vertices = smpl_output.vertices + diff.view(-1, 1, 3)
        return smpl_output

    def _preprocess_param(self, param: np.ndarray) -> torch.Tensor:
        """Prepare the parameters for SMPL
        """
        if not isinstance(param, torch.Tensor):
            param = torch.tensor(param, dtype=torch.float32)
        param = param.to(self.device).reshape(1, -1)
        return param

    def get_smpl(self, pose_params: np.ndarray, translation_params: np.ndarray) -> np.ndarray:
        """Get the SMPL mesh vertices from the target pose and global translation

        Args:
            pose_params: Pose parameters vector of shape (72)
            translation_params: Global translation vector of shape (3)

        Returns:
            np.ndarray: vertices of SMPL model
        """
        global_orient = self._preprocess_param(pose_params[:3])
        body_pose = self._preprocess_param(pose_params[3:])
        translation_params = self._preprocess_param(translation_params)

        smpl_params = dict(global_orient=global_orient, body_pose=body_pose, transl=translation_params)

        with torch.no_grad():
            output = self.smpl(**smpl_params)
        output = self._center_output(self.smpl, smpl_params, output)
        return output.vertices.squeeze(0).cpu().numpy()



from typing import Dict, Iterable, Union, Sequence

import numpy as np

from . import primitives
from .base import Renderable
from .mesh import Mesh
from .pointcloud import PointCloud
from ..colors import UniformColors
from ..colors.base import Colors, ColorsList
from ..internal import Singleton
from ..internal.types import Vector3d, RotationParams
from ..materials.base import Material, MaterialList


class RenderablesCollection(metaclass=Singleton):
    def __init__(self):
        self._renderables: Dict[str, Renderable] = dict()

    # =================================================== PointCloud ===================================================
    def add_pointcloud(
            self,
            vertices: np.ndarray,
            material: Material,
            colors: Colors,
            point_size: float = 0.006,
            base_primitive: str = "CUBE",
            particle_emission_strength: int = 1,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> PointCloud:
        """Add PointCloud object to the scene.
        The object supports uniform (UniformColors) and per-vertex (VertexColors) coloring.

        Args:
            vertices (np.ndarray): point cloud vertices
            material (Material): Material instance
            colors (Colors): VertexColors or UniformColors instance
            point_size (float, optional): size of a primitive, representing each vertex (default: 0.006)
            base_primitive (str, optional): type of primitive for representing each point
                (possible values are PLANE, CUBE, SPHERE, default: CUBE)
            particle_emission_strength (int, optional): strength of the emission from each primitive. This is used to
                increase realism. Values <= 0 turn emission off, values > 0 set the power of emission (default: 1)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created collection in Blender to represent the point cloud.
                If None is passed the tag is automatically generated (default: None)

        Returns:
            PointCloud: created and added to the scene object
        """
        tag = self._process_tag(tag, "PointCloud")
        obj = PointCloud(
            vertices=vertices, material=material, colors=colors, point_size=point_size, base_primitive=base_primitive,
            particle_emission_strength=particle_emission_strength, rotation_mode=rotation_mode, rotation=rotation, translation=translation,
            tag=tag
        )
        self._renderables[tag] = obj
        return obj

    # =============================================== End of PointCloud ================================================

    # ============================================ Mesh and Mesh Primitives ============================================
    def add_mesh(
            self,
            vertices: np.ndarray,
            faces: np.ndarray,
            material: Union[Material, MaterialList],
            colors: Union[Colors, ColorsList],
            faces_material: Sequence[Sequence[int]] = None,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> Mesh:
        """Add Mesh object to the scene.
        The object supports uniform (UniformColors), per-vertex (VertexColors), per-vertex uv (VertexUV),
        per-face uv (FacesUV) and texture (TextureColors and FileTextureColors) coloring.

        Args:
            vertices (np.ndarray): mesh vertices
            faces (np.ndarray): mesh faces
            material (Union[Material, MaterialList]): Material instance or list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or list of Colors instances
            faces_material (Sequence[Sequence[int]], optional): for each face, the material index assigned to it
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            Mesh: created and added to the scene object
        """
        tag = self._process_tag(tag, "Mesh")
        obj = Mesh(
            vertices=vertices, faces=faces, material=material, colors=colors, rotation_mode=rotation_mode,
            rotation=rotation, translation=translation, tag=tag, faces_material=faces_material
        )
        self._renderables[tag] = obj
        return obj

    def add_cube_mesh(
            self,
            size: float,
            material: Union[Material, MaterialList],
            colors: Union[Colors, ColorsList],
            faces_material: Sequence[Sequence[int]] = None,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> primitives.CubeMesh:
        """Add primitives.CircleMesh object to the scene.
        The object supports only uniform coloring (UniformColors).

        Args:
            size (float): size of a primitive in [0, inf]
            material (Union[Material, MaterialList]): Material instance or list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or list of Colors instances
            faces_material (Sequence[Sequence[int]], optional): for each face, the material index assigned to it
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            primitives.CircleMesh: created and added to the scene object
        """
        tag = self._process_tag(tag, "Cube")
        obj = primitives.CubeMesh(
            size=size, material=material, colors=colors, rotation_mode=rotation_mode, rotation=rotation,
            translation=translation, tag=tag, faces_material=faces_material
        )
        self._renderables[tag] = obj
        return obj

    def add_circle_mesh(
            self,
            radius: float,
            material: Union[Material, MaterialList],
            colors: Union[Colors, ColorsList],
            faces_material: Sequence[Sequence[int]] = None,
            num_vertices: int = 32,
            fill_type: str = "NGON",
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> primitives.CircleMesh:
        """Add primitives.CircleMesh object to the scene.
        The object supports only uniform coloring (UniformColors).

        Args:
            radius (float): radius of a primitive in [0, inf]
            material (Union[Material, MaterialList]): Material instance or list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or list of Colors instances
            faces_material (Sequence[Sequence[int]], optional): for each face, the material index assigned to it
            num_vertices (int, optional): number of vertices in primitive in [3, 10000000] (default: 32)
            fill_type (str, optional): fill type, one of [NOTHING, NGON, TRIFAN] (default: NGON)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            primitives.CircleMesh: created and added to the scene object
        """
        tag = self._process_tag(tag, "Circle")
        obj = primitives.CircleMesh(
            radius=radius, material=material, colors=colors, num_vertices=num_vertices, fill_type=fill_type,
            rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag, faces_material=faces_material
        )
        self._renderables[tag] = obj
        return obj

    def add_cylinder_mesh(
            self,
            radius: float,
            height: float,
            material: Union[Material, MaterialList],
            colors: Union[Colors, ColorsList],
            faces_material: Sequence[Sequence[int]] = None,
            num_vertices: int = 32,
            fill_type: str = "NGON",
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> primitives.CylinderMesh:
        """Add primitives.CylinderMesh object to the scene.
        The object supports only uniform coloring (UniformColors).

        Args:
            radius (float): radius of a primitive in [0, inf]
            height (float): height of a primitive in [0, inf]
            material (Union[Material, MaterialList]): Material instance or list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or list of Colors instances
            faces_material (Sequence[Sequence[int]], optional): for each face, the material index assigned to it
            num_vertices (int, optional): number of vertices in primitive in [3, 10000000] (default: 32)
            fill_type (str, optional): fill type, one of [NOTHING, NGON, TRIFAN] (default: NGON)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            primitives.CylinderMesh: created and added to the scene object
        """
        tag = self._process_tag(tag, "Cylinder")
        obj = primitives.CylinderMesh(
            radius=radius, height=height, material=material, colors=colors, num_vertices=num_vertices,
            fill_type=fill_type, rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag, faces_material=faces_material
        )
        self._renderables[tag] = obj
        return obj

    def add_plane_mesh(
            self,
            size: float,
            material: Union[Material, MaterialList],
            colors: Union[Colors, ColorsList],
            faces_material: Sequence[Sequence[int]] = None,
            shadow_catcher: bool = False,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> primitives.PlaneMesh:
        """Add primitives.PlaneMesh object to the scene.
        The object supports only uniform coloring (UniformColors).

        Args:
            size (float): size of a plane in [0, inf]
            material (Union[Material, MaterialList]): Material instance or list of Material instances
            colors (Union[Colors, ColorsList]): Colors instance or list of Colors instances
            faces_material (Sequence[Sequence[int]], optional): for each face, the material index assigned to it
            shadow_catcher (bool, optional): if True, the plane will act as a shadow catcher (default: False)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            primitives.PlaneMesh: created and added to the scene object
        """
        tag = self._process_tag(tag, "Plane")
        obj = primitives.PlaneMesh(
            size=size, material=material, colors=colors, shadow_catcher=shadow_catcher,
            rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag, faces_material=faces_material
        )
        self._renderables[tag] = obj
        return obj

    # ======================================== End of Mesh and Mesh Primitives =========================================

    # ============================================= Parametric Primitives ==============================================
    def add_ellipsoid_nurbs(
            self,
            radius: Vector3d,
            material: Material,
            colors: UniformColors,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> primitives.EllipsoidNURBS:
        """Add primitives.EllipsoidNURBS object to the scene.
        Implemented as NURBS Sphere that is rescaled along axes, supports only uniform coloring (UniformColors).

        Args:
            radius (float): radius of a primitive in [0, inf]
            material (Material): Material instance
            colors (UniformColors): UniformColors instance
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            primitives.EllipsoidNURBS: created and added to the scene object
        """
        tag = self._process_tag(tag, "Ellipsoid")
        obj = primitives.EllipsoidNURBS(
            radius=radius, material=material, colors=colors, rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
        )
        self._renderables[tag] = obj
        return obj

    def add_sphere_nurbs(
            self,
            radius: float,
            material: Material,
            colors: UniformColors,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> primitives.SphereNURBS:
        """Add primitives.SphereNURBS object to the scene.
        The object supports only uniform coloring (UniformColors).

        Args:
            radius (float): radius of a primitive in [0, inf]
            material (Material): Material instance
            colors (UniformColors): UniformColors instance
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            primitives.SphereNURBS: created and added to the scene object
        """
        tag = self._process_tag(tag, "Sphere")
        obj = primitives.SphereNURBS(
            radius=radius, material=material, colors=colors, rotation_mode=rotation_mode, rotation=rotation, translation=translation, tag=tag
        )
        self._renderables[tag] = obj
        return obj

    def add_curve_nurbs(
            self,
            keypoints: np.ndarray,
            radius: float,
            material: Material,
            colors: UniformColors,
            rotation_mode: str = "quaternionWXYZ",
            rotation: RotationParams = None,
            translation: Vector3d = (0, 0, 0),
            tag: str = None
    ) -> primitives.CurveBezier:
        """Add primitives.CurveBezier object to the scene.
        Keypoints are intermediate points of the Bezier Curve. The object supports
        only uniform coloring (UniformColors).

        Args:
            keypoints (np.ndarray): keypoints for the curve
            radius (float): radius of a primitive in [0, inf]
            material (Material): Material instance
            colors (UniformColors): UniformColors instance
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str, optional): name of the created object in Blender. If None is passed, the tag
                is automatically generated (default: None)

        Returns:
            primitives.CurveBezier: created and added to the scene object
        """
        tag = self._process_tag(tag, "Curve")
        obj = primitives.CurveBezier(
            keypoints=keypoints, radius=radius, material=material, colors=colors, rotation_mode=rotation_mode, rotation=rotation,
            translation=translation, tag=tag
        )
        self._renderables[tag] = obj
        return obj

    # ========================================== End of Parametric Primitives ==========================================

    def _process_tag(self, tag: str, default_prefix: str = "Renderable"):
        renderable_keys = self._renderables.keys()

        if tag is None:
            _tag = default_prefix + "_{:03d}"
            index = 0
            while _tag.format(index) in renderable_keys:
                index += 1
            tag = _tag.format(index)
        elif tag in renderable_keys:
            raise RuntimeError(f"Object with tag {tag} is already in collection.")

        return tag

    def keys(self):
        return self._renderables.keys()

    def values(self):
        return self._renderables.values()

    def items(self):
        return self._renderables.items()

    def remove(self, obj_or_tag: Union[Renderable, str]):
        assert isinstance(obj_or_tag, (Renderable, str)), "Only Renderable object or it's tag is allowed"
        if isinstance(obj_or_tag, str):
            tag = obj_or_tag
        else:
            obj = obj_or_tag
            tag = obj.tag
        self.__delitem__(tag)

    def __getitem__(self, key: str) -> Renderable:
        return self._renderables[key]

    def __delitem__(self, key: str):
        self._renderables[key]._blender_remove_object()
        del self._renderables[key]

    def __iter__(self) -> Iterable:
        return iter(self._renderables)

    def __len__(self) -> int:
        return len(self._renderables)

    def _reset(self):
        self._renderables = dict()


from abc import abstractmethod

import numpy as np

from .base import Light
from ..internal.types import Vector3d, Vector2d


class AreaLight(Light):
    """Base class for different AreaLights varying in shape.
    """

    @abstractmethod
    def __init__(
            self,
            color: Vector3d,
            strength: float,
            tag: str,
            cast_shadows: bool = True,
            **kwargs
    ):
        """Creates AreaLight light source in Blender. The method is called from child classes

        Args:
            color (Vector3d): color of the light source
            strength (float): strength of the light source emitted over the entire area of the light in all directions
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        blender_light = self._blender_create_light(tag, "AREA")
        super().__init__(**kwargs, tag=tag, blender_object=blender_light)
        self.color = color
        self.strength = strength
        self.cast_shadows = cast_shadows


class SquareAreaLight(AreaLight):
    def __init__(
            self,
            size: float,
            **kwargs
    ):
        """Creates SquareAreaLight light source in Blender through constructor of the parent class AreaLight

        Args:
            size (float): size of the area of the area light
            strength (float): strength of the light source
            color (Vector3d): color of the light source
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)
        self.blender_light.data.shape = "SQUARE"
        self.size = size

    @property
    def size(self) -> float:
        return self.blender_light.data.size

    @size.setter
    def size(self, val: float):
        self.blender_light.data.size = val


class CircleAreaLight(AreaLight):
    def __init__(
            self,
            size: float,
            **kwargs
    ):
        """Creates CircleAreaLight light source in Blender through constructor of the parent class AreaLight

        Args:
            size (float): size of the area of the area light
            strength (float): strength of the light source emitted over the entire area of the light in all directions
            color (Vector3d): color of the light source
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)
        self.blender_light.data.shape = "DISK"
        self.size = size

    @property
    def size(self) -> float:
        return self.blender_light.data.size

    @size.setter
    def size(self, val: float):
        self.blender_light.data.size = val


class RectangleAreaLight(AreaLight):
    def __init__(
            self,
            size: Vector2d,
            **kwargs
    ):
        """Creates RectangleAreaLight light source in Blender through constructor of the parent class AreaLight

        Args:
            size (Vector2d): [x, y] sizes of the area light
            strength (float): strength of the light source emitted over the entire area of the light in all directions
            color (Vector3d): color of the light source
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)
        self.blender_light.data.shape = "RECTANGLE"
        self.size = size

    @property
    def size(self) -> np.ndarray:
        return np.array([self.blender_light.data.size, self.blender_light.data.size_y])

    @size.setter
    def size(self, val: Vector2d):
        self.blender_light.data.size = val[0]
        self.blender_light.data.size_y = val[1]


class EllipseAreaLight(AreaLight):
    def __init__(
            self,
            size: Vector2d,
            **kwargs
    ):
        """Creates EllipseAreaLight light source in Blender through constructor of the parent class AreaLight

        Args:
            size (Vector2d): [x, y] sizes of the area light
            strength (float): strength of the light source emitted over the entire area of the light in all directions
            color (Vector3d): color of the light source
            cast_shadows (bool, optional): whether the light source casts shadows or not (default: True)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)
        self.blender_light.data.shape = "ELLIPSE"
        self.size = size

    @property
    def size(self) -> np.ndarray:
        return np.array([self.blender_light.data.size, self.blender_light.data.size_y])

    @size.setter
    def size(self, val: Vector2d):
        self.blender_light.data.size = val[0]
        self.blender_light.data.size_y = val[1]



import contextlib
import random

import blendipose.blendify.cameras
import blendipose.blendify.materials.base
import blendipose.util as util
import bpy
import cameravision
import cv2
import numpy as np
import shapely
import simplepyutils as spu
import trimesh
import framepump
from typing import List
from blendipose.blendify import scene
from blendipose.blendify.colors import TextureColorsViaTempFile, UniformColors, VertexUV
from blendipose.blendify.materials import PrincipledBSDFMaterial, PrincipledBSDFWireframeMaterial


class Renderer(contextlib.AbstractContextManager):
    def __init__(
        self,
        resolution,
        body_model_faces,
        out_video_path=None,
        out_fps=None,
        audio_path=None,
        show_image=True,
        num_views=1,
        frame_background=False,
        show_ground_plane=True,
        ground_plane_height=1000,
        world_up=(0, -1, 0),
        show_camera=True,
        preview_resolution=None,
        body_alpha=0.75,
        motion_blur=False,
        depth_of_field=False,
        adaptive_sampling_threshold=1,
        samples=16,
    ):
        self.viz = preview_resolution is not None

        cycles = bpy.context.scene.cycles
        # cycles.use_adaptive_sampling = True
        # cycles.adaptive_threshold = 10
        # cycles.min_samples = 1
        self.motion_blur = motion_blur

        render = bpy.context.scene.render
        if out_fps:
            render.fps = int(round(out_fps))
            render.fps_base = render.fps / out_fps
        if motion_blur:
            render.use_motion_blur = True
            render.motion_blur_shutter = 0.5
            cycles.motion_blur_position = 'CENTER'

        util.set_world_up(world_up)

        self.adaptive_sampling_threshold = adaptive_sampling_threshold
        self.samples = samples

        self.preview_resolution = (
            preview_resolution if preview_resolution is not None else resolution
        )

        self.show_ground = show_ground_plane
        self.show_camera = show_camera
        self.depth_of_field = depth_of_field
        self.ground = self._add_ground(ground_plane_height / 1000) if self.show_ground else None
        self._add_lights()
        if show_camera:
            self.camera_displays = [CameraDisplay(show_image=show_image) for _ in range(num_views)]
        else:
            self.camera_displays = []
        self.show_image = show_image
        self.body_meshes = BodyMeshCollection(body_model_faces, body_alpha)

        self.resolution = resolution
        self.fps = out_fps
        self.stencil_mesh = None
        self.distort_image = lambda x: x
        self.view_camera = None
        self.prev_undist_cam = None

        self.set_view_camera(cameravision.Camera.from_fov(55, resolution))
        self.frame_background = frame_background
        self.video_writer = framepump.VideoWriter(out_video_path, out_fps, audio_path)

    def new_sequence_output(self, out_video_path, fps, audio_source_path=None):
        self.video_writer.start_sequence(
            out_video_path, fps=fps, audio_source_path=audio_source_path
        )
        bpy.context.scene.frame_set(0)
        scene._frame_number = 0
        bpy.context.scene.render.fps = int(round(fps))
        bpy.context.scene.render.fps_base = bpy.context.scene.render.fps / fps
        self.body_meshes.prev_vertices = None
        self.fps = fps

    def update(
        self,
        frame=None,
        mask=None,
        boxes=(),
        poses=(),
        camera=None,
        poses_true=(),
        poses_alt=(),
        vertices=(),
        vertices_true=(),
        vertices_alt=(),
        viz_camera=None,
        viz_imshape=None,
        uncerts=None,
        uncerts_alt=None,
        colors=None,
    ):
        if len(boxes) == 0:
            boxes = np.zeros((0, 4), np.float32)

        if uncerts is not None:
            vertices = [
                np.concatenate([v, uncert[..., np.newaxis]], axis=-1)
                for v, uncert in zip(vertices, uncerts)
            ]
        if uncerts_alt is not None:
            vertices_alt = [
                np.concatenate([v, uncert[..., np.newaxis]], axis=-1)
                for v, uncert in zip(vertices_alt, uncerts_alt)
            ]

        self.body_meshes.update(vertices, self.fps, colors=colors)

        if self.show_camera:
            self.camera_displays[0].update(camera, frame)

        if self.depth_of_field:
            if self.show_image and np.linalg.norm(viz_camera.t - camera.t) < 500:
                focus_distance = None
            elif len(vertices) > 0:
                focus_distance = np.linalg.norm(np.mean(vertices[0], axis=0) - viz_camera.t) / 1000
            else:
                focus_distance = 1.0

        else:
            focus_distance = None

        self.set_view_camera(viz_camera, focus_distance=focus_distance)
        rendered_im = self.render_image()
        if mask is not None:
            # current_mask = rendered_im[:, :, 3]
            # current_mask == 1.0
            rendered_im[:, :, 3] *= mask

        if self.frame_background:
            frame = resize_image(frame, rendered_im.shape, padding_value=65535)
            im_with_bg = util.alpha_blend(rendered_im, frame, dtype=np.uint16)
        else:
            im_with_bg = util.alpha_blend(rendered_im, 65535, dtype=np.uint16)

        if self.viz:
            im_show = util.resize_image(im_with_bg, self.preview_resolution[::-1])
            cv2.imshow('img', cv2.cvtColor(im_show, cv2.COLOR_RGB2BGR))
            cv2.waitKey(1)

        if self.video_writer.is_active():
            self.video_writer.append_data(im_with_bg)

        return rendered_im, im_with_bg

    def render_image(self):
        bpy.context.scene.cycles.seed = random.randint(0, 10_000)
        undistorted_image = scene.render(
            use_denoiser=True,
            adaptive_sampling_threshold=self.adaptive_sampling_threshold,
            samples=self.samples,
            color_dtype=np.uint16,
        )
        return self.distort_image(undistorted_image)

    def set_view_camera(self, cameravision_camera, focus_distance=None):
        if False and self.view_camera is not None and self.view_camera.is_equal(cameravision_camera):
            return

        if self.stencil_mesh is not None:
            scene.renderables.remove(self.stencil_mesh)
            self.stencil_mesh = None

        c = cameravision_camera.copy()
        if c.has_distortion():
            undist_cam, box, poly = c.undistort_precise(
                alpha_balance=1.0, imshape_distorted=self.resolution[::-1], inplace=False
            )
            imshape_undistorted = np.ceil(box[2:][::-1]).astype(int)

            def distort_fn(image):
                return cameravision.reproject_image(
                    image,
                    undist_cam,
                    c,
                    output_imshape=self.resolution[::-1],
                    cache_maps=True,
                    use_linear_srgb=True,
                    antialias_factor=2,
                    interp=cv2.INTER_CUBIC,
                )[0]

            self.distort_image = distort_fn
            self._add_holdout_stencil(undist_cam, poly)
        else:
            undist_cam = c
            self.distort_image = lambda x: x
            imshape_undistorted = self.resolution[::-1]

        set_cameravision_camera(
            scene,
            undist_cam,
            imshape_undistorted[:2][::-1],
            near=0.1,
            far=10000,
            focus_distance=focus_distance,
        )

        if self.motion_blur:
            insert_camera_keyframe(scene.camera, bpy.context.scene.frame_current + 1)
            if self.prev_undist_cam is not None:
                undist_cam_extrap = extrapolate(self.prev_undist_cam, undist_cam)
                set_cameravision_camera(
                    scene, undist_cam_extrap, imshape_undistorted[:2][::-1], near=0.1, far=10000
                )
                insert_camera_keyframe(scene.camera, bpy.context.scene.frame_current + 2)
            self.prev_undist_cam = undist_cam

        self.view_camera = cameravision_camera.copy()

    def _add_holdout_stencil(self, camera, polygon):
        # Add a flat rectangle in front of the camera, with a cutout for the given polygon
        # ie, the polygon is a hole in the rectangle, and the rectangle has a holdout shader
        # material
        # This is useful for avoiding rendering unnecessary parts of the scene

        stencil_mesh = poly_to_stencil_trimesh(polygon)
        vs = camera.image_to_world(stencil_mesh.vertices[:, :2], camera_depth=0.101)
        self.stencil_mesh = scene.renderables.add_mesh(
            util.world_to_blender(vs),
            stencil_mesh.faces,
            material=HoldoutMaterial(),
            colors=UniformColors((0, 0, 0, 0)),
        )
        self.stencil_mesh.emit_shadows = False
        self.stencil_mesh._blender_mesh.is_holdout = True

    def _add_ground(self, height):
        texture = np.zeros((20, 20, 3), np.uint8)
        texture[::2] = 255 - texture[::2]
        texture[:, ::2] = 255 - texture[:, ::2]
        texture[texture == 255] = 155
        texture[texture == 0] = 100

        # up normal is the normal vector of the plane
        # height is the dot product of the points on the plane and the normal vector
        mesh = util.add_billboard(
            texture=texture,
            scale=20,
            translation=(0, 0, height),
            material=PrincipledBSDFMaterial(specular=0.0, alpha=0.6, roughness=1, metallic=0.0),
        )

        return mesh

    def _add_lights(self):
        strengths = np.array([0.9 * 2, 0.9 * 2, 0.7 * 2, 0.9 * 0.7, 0.9 * 0.7, 0.7 * 0.7])
        # by default the sun shines towards the negative z-axis (ie, downwards)
        rotations = [
            (10, 45, 0),
            (145, 25, 0),
            (-125, 35, 0),
            (10, 180 + 45, 0),
            (145, 180 + 25, 0),
            (-125, 180 + 35, 0),
        ]

        for strength, rotation in zip(strengths, rotations):
            scene.lights.add_sun(
                strength=strength,
                angular_diameter=0.1,
                rotation=rotation,
                rotation_mode='eulerZXY',
                cast_shadows=True,
            )
        scene.lights.set_background_light(strength=0.35, color=(1.0, 1.0, 1.0))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.video_writer.close()


def insert_camera_keyframe(camera, frame):
    camera = camera.blender_camera
    camera.keyframe_insert(data_path="location", frame=frame)
    camera.keyframe_insert(data_path="rotation_quaternion", frame=frame)
    camera.data.keyframe_insert(data_path="lens", frame=frame)
    camera.data.keyframe_insert(data_path="shift_x", frame=frame)
    camera.data.keyframe_insert(data_path="shift_y", frame=frame)
    for fcurve in [
        *camera.animation_data.action.fcurves,
        *camera.data.animation_data.action.fcurves,
    ]:
        fcurve.keyframe_points[-1].interpolation = 'LINEAR'


class BodyMeshCollection:
    def __init__(self, faces, alpha, color=None, motion_blur=True):
        self.faces = faces
        self.meshes: List[blendipose.blendify.renderables.Mesh] = []
        color = (0 / 255, 100 / 255, 255 / 255, alpha)
        self.material = PrincipledBSDFMaterial(
            specular=1,
            metallic=0.25,
            roughness=0.54,
            emission_strength=0,  # 0.08,
            alpha=alpha,
            emission=color,
        )
        self.colors = UniformColors(color)

        self.prev_vertices = None
        self.motion_blur = motion_blur

    def _create_mesh(self, vertices, color=None):
        if color is None:
            colors = self.colors
        else:
            colors = UniformColors(color)

        mesh = scene.renderables.add_mesh(
            vertices=util.world_to_blender(vertices),
            faces=self.faces,
            material=self.material,
            colors=colors,
        )
        mesh.set_smooth()

        if self.motion_blur:
            mesh._blender_mesh.attributes.new(name='velocity', type='FLOAT_VECTOR', domain='POINT')
        return mesh

    def update(self, vertices, fps, colors=None):
        if colors is not None:
            for m in self.meshes:
                scene.renderables.remove(m)
            del self.meshes[:]
            self.meshes += [self._create_mesh(v, c) for v, c in zip(vertices, colors)]
            return

        n_people_new = len(vertices)
        n_people_old = len(self.meshes)

        if self.motion_blur:
            if self.prev_vertices is not None:
                velocities = [
                    (v - prev_v) * fps for v, prev_v in zip(vertices, self.prev_vertices)
                ]
                for m, vel in zip(self.meshes, velocities):
                    update_mesh_velocities(m, util.world_to_blender(vel))
            self.prev_vertices = vertices

        for m, v in zip(self.meshes, vertices):
            m.update_vertices(util.world_to_blender(v))

        if n_people_new < n_people_old:
            for m in self.meshes[n_people_new:]:
                scene.renderables.remove(m)
            del self.meshes[n_people_new:]
        elif n_people_new > n_people_old:
            self.meshes += [self._create_mesh(v) for v in vertices[n_people_old:]]


def update_mesh_velocities(mesh, velocities):
    for attr, vel in zip(mesh._blender_mesh.attributes['velocity'].data, velocities):
        attr.vector = vel


class CameraDisplay:
    def __init__(self, camera_depth=500, show_image=True, add_campoint=True):
        self.camera_depth = camera_depth
        self.add_campoint = add_campoint
        self.show_image = show_image
        self.texture_colors = None
        self.image_mesh = None
        self.camera_mesh = None
        self.imshape = None
        self.is_initialized = False
        self.uv_map = None

    def _initial_update(self, camera, frame):
        self.imshape = frame.shape
        h, w = frame.shape[:2]
        image_corners = np.array([[0, 0], [w - 1, 0], [w - 1, h - 1], [0, h - 1]], np.float32)
        image_corners_world = camera.image_to_world(image_corners, camera_depth=self.camera_depth)
        if self.add_campoint:
            vertices = np.array([camera.t, *image_corners_world])
            faces = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 1]])
        else:
            vertices = np.array([camera.t - camera.R[2] * 0.5, *image_corners_world])
            faces = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 1]])

        vertices = util.world_to_blender(vertices)

        self.camera_mesh = scene.renderables.add_mesh(
            vertices,
            faces,
            material=PrincipledBSDFWireframeMaterial(),
            colors=UniformColors((0, 0, 0, 0)),
        )
        self.camera_mesh.emit_shadows = False
        self.uv_map = VertexUV(np.array([[0, 1], [1, 1], [1, 0], [0, 0]]))
        self.texture_colors = TextureColorsViaTempFile(frame, self.uv_map, has_alpha=False)
        if self.show_image:
            self.image_mesh = scene.renderables.add_mesh(
                vertices[1:],
                faces[:2],
                colors=self.texture_colors,
                material=PrincipledBSDFMaterial(
                    specular=0.0,
                    alpha=0.5,
                    roughness=1,
                    metallic=0.0,
                    emission_strength=1,
                    base_color=(0, 0, 0, 1),
                    use_colors_for_emission=True,
                ),
            )
            self.image_mesh.emit_shadows = False

        self.is_initialized = True

    def update(self, camera, frame):
        camera = camera.copy()

        if not self.is_initialized:
            self._initial_update(camera, frame)

        h, w = frame.shape[:2]
        image_corners = np.array([[0, 0], [w - 1, 0], [w - 1, h - 1], [0, h - 1]], np.float32)
        image_corners_world = camera.image_to_world(image_corners, camera_depth=self.camera_depth)
        vertices = np.array([camera.t, *image_corners_world])
        vertices = util.world_to_blender(vertices)
        self.camera_mesh.update_vertices(vertices)
        if self.image_mesh is not None:
            self.image_mesh.update_vertices(vertices[1:])

            if frame.shape != self.imshape:
                self.imshape = frame.shape
                self.texture_colors = TextureColorsViaTempFile(frame, self.uv_map)
                self.image_mesh.update_colors(self.texture_colors)
            else:
                self.texture_colors.update_pixels(frame)


def set_cameravision_camera(
    scene,
    camera,
    resolution,
    near=0.1,
    far=100.0,
    tag='camera',
    resolution_percentage=100,
    focus_distance=None,
):
    # To go from blender world to blender cam, we go from
    # 1) blender world to world (util.WORLD_TO_BLENDERWORLD_ROTATION_MAT.T),
    # 2) world to cam (camera.R),
    # 3) cam to blender cam (util.CAM_TO_BLENDERCAM_ROTATION_MAT)

    R = util.CAM_TO_BLENDERCAM_ROTATION_MAT @ camera.R @ util.WORLD_TO_BLENDERWORLD_ROTATION_MAT.T
    t = util.world_to_blender(camera.t)
    f = camera.intrinsic_matrix[0, 0]
    center = camera.intrinsic_matrix[:2, 2] / resolution

    blendify_cam = scene.camera
    if isinstance(blendify_cam, blendipose.blendify.cameras.PerspectiveCamera) and np.all(
        blendify_cam.resolution == resolution
    ):
        # R is the rotation matrix from the world to camera, so we must transpose it
        # because this function expects the transformation from camera to world
        blendify_cam.set_position(translation=t, rotation=R.T, rotation_mode='rotmat')
        blendify_cam.focal_dist = f
        blendify_cam.center = center
        blendify_cam.far = far
        blendify_cam.near = near
    else:
        scene.set_perspective_camera(
            resolution,
            rotation=R.T,
            rotation_mode='rotmat',
            translation=t,
            focal_dist=f,
            center=center,
            near=near,
            far=far,
            tag=tag,
            resolution_percentage=resolution_percentage,
        )

    if focus_distance is not None:
        c = scene.camera.blender_camera
        c.data.dof.use_dof = True
        c.data.dof.focus_distance = focus_distance
        c.data.dof.aperture_fstop = 30
        c.data.dof.aperture_blades = 3


def extrapolate(cam1, cam_mid):
    t2 = 2 * cam_mid.t - cam1.t
    R2 = cam_mid.R @ cam1.R.T @ cam_mid.R

    center1 = cam1.intrinsic_matrix[:2, 2]
    center_mid = cam_mid.intrinsic_matrix[:2, 2]
    center2 = 2 * center_mid - center1

    f1 = cam1.intrinsic_matrix[0, 0]
    fmid = cam_mid.intrinsic_matrix[0, 0]
    f2 = 2 * fmid - f1

    cam2 = cam_mid.copy()
    cam2.t = t2
    cam2.R = R2

    cam2.intrinsic_matrix[0, 0] = f2
    cam2.intrinsic_matrix[1, 1] = f2
    cam2.intrinsic_matrix[:2, 2] = center2
    return cam2


def resize_image(im, dst_shape, padding_value=255):
    imshape = im.shape[:2]
    dst_shape = dst_shape[:2]
    if im.shape == dst_shape:
        return im

    interp = cv2.INTER_LINEAR if dst_shape[0] > imshape[0] else cv2.INTER_AREA
    factor = np.min(np.array(dst_shape, dtype=np.float32) / np.array(imshape, dtype=np.float32))
    dst_shape2 = spu.rounded_int_tuple(np.array(imshape, dtype=np.float32) * factor)
    resized = cv2.resize(im, (dst_shape2[1], dst_shape2[0]), interpolation=interp)
    # now pad
    padding_total = np.array(dst_shape) - np.array(dst_shape2)
    padding_top = padding_total[0] // 2
    padding_bottom = padding_total[0] - padding_top
    padding_left = padding_total[1] // 2
    padding_right = padding_total[1] - padding_left
    padded = cv2.copyMakeBorder(
        resized,
        padding_top,
        padding_bottom,
        padding_left,
        padding_right,
        cv2.BORDER_CONSTANT,
        value=padding_value,
    )
    return padded


def polygon_to_trimesh(polygon: shapely.Polygon):
    verts2d, faces = trimesh.creation.triangulate_polygon(polygon)
    verts = np.column_stack([verts2d, np.ones(len(verts2d))])
    return trimesh.Trimesh(vertices=verts, faces=faces)


def multipolygon_to_trimesh(multipolygon: shapely.MultiPolygon):
    return trimesh.util.concatenate([polygon_to_trimesh(p) for p in multipolygon.geoms])


def poly_to_stencil_trimesh(poly: shapely.Polygon):
    polybox = shapely.box(*poly.bounds)
    stencilpoly = polybox - poly.buffer(5)
    if isinstance(stencilpoly, shapely.Polygon):
        return polygon_to_trimesh(stencilpoly)
    else:
        return multipolygon_to_trimesh(stencilpoly)


class HoldoutMaterial(blendipose.blendify.materials.base.Material):
    """A class which manages a Holdout Blender material."""

    def __init__(self, use_backface_culling=True):
        super().__init__(use_backface_culling=use_backface_culling)

    def create_material(
        self, name: str = "holdout_material"
    ) -> blendipose.blendify.materials.base.MaterialInstance:
        object_material = bpy.data.materials.new(name=name)
        object_material.use_nodes = True
        object_material.use_backface_culling = self._use_backface_culling

        nodes = object_material.node_tree.nodes
        links = object_material.node_tree.links

        # Remove the default Principled BSDF node if it exists
        # if "Principled BSDF" in nodes:
        #    nodes.remove(nodes["Principled BSDF"])

        # Create the Holdout node
        holdout_node = nodes.new("ShaderNodeHoldout")
        holdout_node.location = (0, 0)

        # Link the Holdout node's output to the Material Output's Surface input
        material_output = nodes.get("Material Output")
        links.new(holdout_node.outputs[0], material_output.inputs["Surface"])

        material_instance = blendipose.blendify.materials.base.MaterialInstance(
            blender_material=object_material,
            inputs={
                'Color': nodes["Principled BSDF"].inputs["Base Color"],
                'Alpha': nodes["Principled BSDF"].inputs["Alpha"],
            },
        )
        return material_instance



import bpy  # imported first to prepare internal Blender environment
from . import renderables
from .scene import Scene

__all__ = [
    "scene",
    "__version__"
]

VERSION = (2, 0, 0)  # PEP 386
__version__ = ".".join([str(x) for x in VERSION])

scene = Scene()


def _get_scene():
    return scene



from abc import ABC, abstractmethod

import bpy
import numpy as np

from .base import Colors, ColorsMetadata
from ..internal.texture import _copy_values_to_image
from memory_tempfile import MemoryTempfile
import cv2

class UVMap(ABC):
    """An abstract container template for storing a UV coordinate map
    """

    @abstractmethod
    def __init__(self, data: np.ndarray):
        self._data = data

    @property
    def data(self) -> np.ndarray:
        """Get the stored UV map data

        Returns:
            np.ndarray: UV map data
        """
        return self._data


class VertexUV(UVMap):
    """A container which stores a UV coordinate for every vertex
    In the form of (N,2) array (N vertices, 2 UV coordinates for each)
    """

    def __init__(self, data: np.ndarray):
        super().__init__(data)


class FacesUV(UVMap):
    """A container which stores a UV coordinate for every vertex in every triangle face
    In the form of (M,3,2) array (M faces, 3 vertices in each face, 2 UV coordinates for each
    vertex in triangle)
    """

    def __init__(self, data: np.ndarray):
        super().__init__(data)


class UVColors(Colors):
    """An abstract container for storing color information bound to UV coordinate space
    """

    @abstractmethod
    def __init__(self, uv_map: UVMap):
        super().__init__()
        self._uv_map = uv_map

    @property
    def uv_map(self) -> UVMap:
        """Get the stored UV map

        Returns:
            UVMap: stored UV map
        """
        return self._uv_map


class TextureColors(UVColors):
    """A container which stores texture in form of pixels array and the corresponding UV mapping
    """

    def __init__(self, texture: np.ndarray, uv_map: UVMap, interpolation: str = 'Linear'):
        """Create the texture container and initialize a Blender texture with the pixels data

        Args:
            texture (np.ndarray): pixels data
            uv_map (UVMap): corresponding UV map
            interpolation (str, optional): interpolation method for the texture, values can be
            'Linear', 'Closest', 'Cubic' or 'Smart'. Defaults to 'Linear'.
        """
        assert texture.ndim == 3 and texture.shape[2] in (
        3, 4), "Texture must be a 3D array of shape (H, W, 3) or (H, W, 4)"
        assert interpolation in ('Linear', 'Closest', 'Cubic',
                                 'Smart'), ("Interpolation must be one of 'Linear', 'Closest', "
                                            "'Cubic' or 'Smart'")
        super().__init__(uv_map)
        if texture.dtype == np.uint8:
            texture = texture.astype(np.float32) / 255.
        blender_image = bpy.data.images.new(name="tex_image", width=texture.shape[1],
                                            height=texture.shape[0], alpha=False)
        last_dim = texture.shape[2]
        # Flip the data vertically before copying
        texture = texture[::-1]
        if last_dim == 3:
            texture = np.concatenate([texture, np.ones_like(texture[..., :1])], axis=-1)
        blender_image.pixels.foreach_set(texture.ravel())
        blender_image.pack()
        self._texture = blender_image
        self._metadata = ColorsMetadata(
            type=self.__class__,
            has_alpha=last_dim == 4,
            color=None,
            texture=self._texture,
            interpolation=interpolation
        )

    @property
    def blender_texture(self) -> bpy.types.Image:
        """Get the current Blender texture created from the pixels array

        Returns:
            bpy.types.Image: current Blender texture
        """
        return self._texture

    def update_pixels(self, texture: np.ndarray):
        """Update the texture with new pixel data

        Args:
            texture (np.ndarray): new pixel data
        """
        assert texture.ndim == 3 and texture.shape[2] in (3, 4), "Texture must be a 3D array of shape (H, W, 3) or (H, W, 4)"
        assert (texture.shape[0] == self._texture.size[1] and
                texture.shape[1] == self._texture.size[0]), "Texture shape must match the original texture shape"
        if texture.dtype == np.uint8:
            texture = texture.astype(np.float32) / 255.
        last_dim = texture.shape[2]
        # Flip the data vertically before copying
        texture = texture[::-1]
        if last_dim == 3:
            texture = np.concatenate([texture, np.ones_like(texture[..., :1])], axis=-1)
        self._texture.pixels.foreach_set(texture.ravel())
        #self._texture.pixels = texture.ravel()
        self._texture.update()


class FileTextureColors(UVColors):
    """A container which stores path to the texture file and the corresponding UV mapping
    """

    def __init__(self, texture_path: str, uv_map: UVMap, interpolation: str = 'Linear', has_alpha: bool=True):
        """Create the texture container and load the texture from the path as a Blender texture

        Args:
            texture_path (str): path to the texture
            uv_map: corresponding UV map
            interpolation (str, optional): interpolation method for the texture, values can be
            'Linear', 'Closest', 'Cubic' or 'Smart'. Defaults to 'Linear'.
            has_alpha (bool, optional): whether the texture's alpha should be used. Defaults to True.
        """
        super().__init__(uv_map)
        self._texture = bpy.data.images.load(texture_path)
        self._metadata = ColorsMetadata(
            type=self.__class__,
            has_alpha=has_alpha,
            color=None,
            texture=self._texture,
            interpolation=interpolation
        )

    @property
    def blender_texture(self) -> bpy.types.Image:
        """Get the current Blender texture created from the pixels array

        Returns:
            bpy.types.Image: current Blender texture
        """
        return self._texture



class VideoFileTextureColors(UVColors):
    """A container which stores path to the texture file and the corresponding UV mapping
    """

    def __init__(self, texture_path: str, uv_map: UVMap, interpolation: str = 'Linear', frame_start: int = 0):
        """Create the texture container and load the texture from the path as a Blender texture

        Args:
            texture_path (str): path to the texture
            uv_map: corresponding UV map
            interpolation (str, optional): interpolation method for the texture, values can be
            'Linear', 'Closest', 'Cubic' or 'Smart'. Defaults to 'Linear'.
        """
        super().__init__(uv_map)
        self._texture = bpy.data.images.load(texture_path)
        self._texture.source = 'MOVIE'
        self._metadata = ColorsMetadata(
            type=self.__class__,
            has_alpha=False,
            color=None,
            texture=self._texture,
            interpolation=interpolation,
            frame_start=frame_start
        )


    @property
    def blender_texture(self) -> bpy.types.Image:
        """Get the current Blender texture created from the pixels array

        Returns:
            bpy.types.Image: current Blender texture
        """
        return self._texture


class TextureColorsViaTempFile(FileTextureColors):
    def __init__(
            self, texture: np.ndarray, uv_map: UVMap, interpolation: str = 'Linear', has_alpha: bool = True):
        self.tempfile_maker = MemoryTempfile()
        self.tempfile = self.tempfile_maker.NamedTemporaryFile(mode='wb', suffix='.png')
        self.update_pixels(texture)
        super().__init__(self.tempfile.name, uv_map, interpolation, has_alpha)


    def update_pixels(self, texture: np.ndarray):
        _, imbuf = cv2.imencode('.png', cv2.cvtColor(texture, cv2.COLOR_RGB2BGR))
        self.tempfile.seek(0)
        self.tempfile.write(memoryview(imbuf))
        self.tempfile.truncate()
        self.tempfile.flush()
        #self._texture.reload() # apparently this is not necessary?



import numpy as np

from ..internal.types import Vector3d


def blend_with_background(img: np.ndarray, bkg_color: Vector3d = (1., 1., 1.)) -> np.ndarray:
    """Blend the RGBA image with uniform colored background, return RGB image

    Args:
        img: RGBA foreground image
        bkg_color: RGB uniform background color (default is white)

    Returns:
        np.ndarray: RGB image blended with background
    """
    bkg_color = np.array(bkg_color)
    if img.dtype == np.uint8:
        bkg_color_uint8 = (bkg_color*255).astype(np.uint8)
        alpha = img[:, :, 3:4].astype(np.int32)
        img_with_bkg = ((img[:, :, :3] * alpha + bkg_color_uint8[None, None, :] * (255 - alpha)) // 255).astype(np.uint8)
    else:
        alpha = img[:, :, 3:4]
        img_with_bkg = (img[:, :, :3] * alpha + bkg_color[None, None, :] * (1. - alpha))
    return img_with_bkg



from typing import Tuple

import bpy

from .base import Material


class WireframeMaterial(Material):
    def __init__(
        self, wireframe_thickness=0.01, wireframe_color=(0., 0., 0., 1.), **kwargs
    ):
        super().__init__(**kwargs)

        self._wireframe_thickness = wireframe_thickness
        self._wireframe_color = wireframe_color

    def overlay_wireframe(self, object_material, base_material_node):
        material_nodes = object_material.node_tree.nodes

        # Create material for wireframe
        wireframe_color_node = material_nodes.new("ShaderNodeRGB")
        wireframe_color_node.outputs["Color"].default_value = self._wireframe_color
        bc_node = material_nodes.new("ShaderNodeBrightContrast")
        object_material.node_tree.links.new(bc_node.inputs["Color"],
                                            wireframe_color_node.outputs["Color"])
        diffuse_node = material_nodes.new("ShaderNodeBsdfDiffuse")
        object_material.node_tree.links.new(diffuse_node.inputs["Color"],
                                            bc_node.outputs["Color"])

        # Create wireframe and mix it with base material
        wireframe_node = material_nodes.new("ShaderNodeWireframe")
        wireframe_node.inputs[0].default_value = self._wireframe_thickness  # thickness
        mix_node = material_nodes.new("ShaderNodeMixShader")

        object_material.node_tree.links.new(mix_node.inputs["Fac"],
                                            wireframe_node.outputs["Fac"])
        object_material.node_tree.links.new(mix_node.inputs[1],
                                            base_material_node.outputs[0])
        object_material.node_tree.links.new(mix_node.inputs[2],
                                            diffuse_node.outputs[0])

        object_material.node_tree.links.new(material_nodes["Material Output"].inputs["Surface"],
                                            mix_node.outputs[0])



from abc import ABC, abstractmethod
from typing import Sequence, Union, Dict, Optional
from dataclasses import dataclass

import bpy


@dataclass
class MaterialInstance:
    blender_material: bpy.types.Material
    inputs: Dict[str, bpy.types.NodeSocket]
    colors_node: Optional[bpy.types.ShaderNode] = None


class Material(ABC):
    def __init__(self, use_backface_culling=False):
        self._use_backface_culling = use_backface_culling

    @abstractmethod
    def create_material(self, name: str = "object_material") -> MaterialInstance:
        pass


MaterialList = Sequence[Material]


def material_property(name: str):
    """Creates a property for the material class to get one of the material parameters

    Args:
        name (str): property name

    Returns:
        property: A class property with defines parameter setting and getting behaviour
    """
    name = "_" + name

    def getter(obj):
        return getattr(obj, name)

    return property(getter)



from abc import abstractmethod

import bpy
import numpy as np

from ..internal.positionable import Positionable
from ..internal.types import Vector3d, RotationParams


class Light(Positionable):
    """Abstract base class for all the light sources.
    """
    @abstractmethod
    def __init__(
        self,
        **kwargs
    ):
        """Passes all arguments to the constructor of the base class

        Args:
            blender_object (bpy.types.Object): instance of Blender Object that is wrapped by the class
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)

    def _blender_create_light(self, tag: str, light_type: str) -> bpy.types.Object:
        light_obj = bpy.data.lights.new(name=tag, type=light_type)
        obj = bpy.data.objects.new(name=tag, object_data=light_obj)

        bpy.context.collection.objects.link(obj)
        return obj

    @property
    def blender_light(self) -> bpy.types.Object:
        return self._blender_object

    @property
    def color(self) -> np.ndarray:
        return np.array(self.blender_light.data.color[:3])

    @color.setter
    def color(self, val: Vector3d):
        val = np.array(val)
        self.blender_light.data.color = val.tolist()

    @property
    def cast_shadows(self) -> bool:
        return self.blender_light.data.cycles.cast_shadow

    @cast_shadows.setter
    def cast_shadows(self, val: bool):
        self.blender_light.data.cycles.cast_shadow = val

    @property
    def strength(self) -> float:
        return self.blender_light.data.energy

    @strength.setter
    def strength(self, val: float):
        self.blender_light.data.energy = val

    @property
    def max_bounces(self) -> int:
        return self.blender_light.data.cycles.max_bounces

    @max_bounces.setter
    def max_bounces(self, val: int):
        self.blender_light.data.cycles.max_bounces = val



from .bsdf import PrincipledBSDFMaterial, PrincipledBSDFWireframeMaterial


class PlasticMaterial(PrincipledBSDFMaterial):
    def __init__(
            self, roughness=0.3, ior=1.45, clearcoat=0.0, specular=0.5,
            clearcoat_roughness=0.0, **kwargs
    ):
        super().__init__(
            roughness=roughness, ior=ior, clearcoat=clearcoat, specular=specular,
            clearcoat_roughness=clearcoat_roughness, **kwargs
        )


class PlasticWireframeMaterial(PrincipledBSDFWireframeMaterial):
    def __init__(
            self, roughness=0.3, ior=1.45, clearcoat=0.0, specular=0.5,
            clearcoat_roughness=0.0, wireframe_thickness=0.01, wireframe_color=(0., 0., 0., 1.),
            **kwargs
    ):
        super().__init__(roughness=roughness, ior=ior, clearcoat=clearcoat, specular=specular,
                         clearcoat_roughness=clearcoat_roughness, wireframe_thickness=wireframe_thickness,
                         wireframe_color=wireframe_color, **kwargs
                         )



import numpy as np

from .base import Camera
from ..internal.types import Vector2d


class PerspectiveCamera(Camera):
    def __init__(
        self,
        focal_dist: float = None,
        fov_x: float = None,
        fov_y: float = None,
        center: Vector2d = None,
        **kwargs
    ):
        """Creates Perspective Camera object in Blender. One of focal_dist, fov_x or fov_y is required to
        set the camera parameters

        Args:
            focal_dist (float, optional): Perspective Camera focal distance in millimeters (default: None)
            fov_x (float, optional): Camera lens horizontal field of view (default: None)
            fov_y (float, optional): Camera lens vertical field of view (default: None)
            center (Vector2d, optional): (x, y), horizontal and vertical shifts of the Camera (default: None)
            resolution (Vector2di): (w, h), the resolution of the resulting image
            near (float, optional): Camera near clipping distance (default: 0.1)
            far (float, optional): Camera far clipping distance (default: 100)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)
        assert not(focal_dist is None and fov_x is None and fov_y is None), \
            "One of focal_dist, fov_x or fov_y is required"
        camera_object = self.blender_camera
        camera_object.data.type = 'PERSP'
        # camera.data.lens_unit = "FOV"
        if focal_dist is not None:
            self.focal_dist = focal_dist
        if center is not None:
            self.center = center
        if fov_x is not None:
            self.fov_x = fov_x
        if fov_y is not None:
            self.fov_y = fov_y

    @property
    def focal_dist(self):
        return self.blender_camera.data.lens

    @focal_dist.setter
    def focal_dist(self, focal: float):
        camera = self.blender_camera
        self._focal_dist = focal
        camera.data.lens = focal

    @property
    def fov_x(self):
        return self.blender_camera.data.angle_x

    @fov_x.setter
    def fov_x(self, val: float):
        self.blender_camera.data.angle_x = val

    @property
    def fov_y(self):
        return self.blender_camera.data.angle_y

    @fov_y.setter
    def fov_y(self, val: float):
        self.blender_camera.data.angle_y = val

    @property
    def center(self) -> np.ndarray:
        camera = self.blender_camera
        ideal_center = (self.resolution - 1.0)/ 2.
        # Blender's camera shift is relative and bounded by [-2, 2]
        center_offset_relative = np.array([camera.data.shift_x, camera.data.shift_y])
        real_center = ideal_center + center_offset_relative * self.resolution
        return real_center

    @center.setter
    def center(self, real_center: Vector2d):
        assert np.all(np.array(real_center) >= -2) and np.all(np.array(real_center) <= 2), \
            ("Blender's camera center is set as a fraction of resolution and "
             "should be in [-2, 2], got {}").format(real_center)
        camera = self.blender_camera
        real_center = np.array(real_center)
        ideal_center = (self.resolution - 1.0) / 2.
        center_offset_relative = real_center - ideal_center / self.resolution
        camera.data.shift_x = center_offset_relative[0]
        camera.data.shift_y = center_offset_relative[1]

    def distance2depth(self, distmap: np.ndarray) -> np.ndarray:
        """Convert map of camera ray lengths (distmap) to map of distances to image plane (depthmap)

        Args:
            distmap (np.ndarray): Distance map

        Returns:
            np.ndarray: Depth map
        """
        img_width, img_height = self.resolution
        cx, cy = self.center
        offsets_x = np.arange(img_width) - cx
        offsets_y = np.arange(img_height) - cy
        grid_offsets_x, grid_offsets_y = np.meshgrid(offsets_x, offsets_y)
        depthmap = np.sqrt(distmap ** 2 / ((grid_offsets_x ** 2 + grid_offsets_y ** 2) / (self.focal_dist ** 2) + 1))
        return depthmap



class OrthographicCamera(Camera):
    def __init__(
        self,
        ortho_scale: float = 1.,
        **kwargs
    ):
        """Creates Orthographic Camera object in Blender

        Args:
            ortho_scale (float, optional): Orthographic Camera scale (similar to zoom) (default: 1.0)
            resolution (Vector2di): (w, h), the resolution of the resulting image
            near (float, optional): Camera near clipping distance (default: 0.1)
            far (float, optional): Camera far clipping distance (default: 100)
            rotation_mode (str): type of rotation representation.
                Can be one of the following:
                - "quaternionWXYZ" - WXYZ quaternion
                - "quaternionXYZW" - XYZW quaternion
                - "rotvec" - axis-angle representation of rotation
                - "rotmat" - 3x3 rotation matrix
                - "euler<mode>" - Euler angles with the specified order of rotation, e.g. XYZ, xyz, ZXZ, etc. Refer to scipy.spatial.transform.Rotation.from_euler for details.
                - "look_at" - look at rotation, the rotation is defined by the point to look at and, optional, the rotation around the forward direction vector (a single float value in tuple or list)
            rotation (RotationParams): rotation parameters according to the rotation_mode
                - for "quaternionWXYZ" and "quaternionXYZW" - Vec4d
                - for "rotvec" - Vec3d
                - for "rotmat" - Mat3x3
                - for "euler<mode>" - Vec3d
                - for "look_at" - Vec3d, Positionable or Tuple[Vec3d/Positionable, float], where float is the rotation around the forward direction vector in degrees
            translation (Vector3d, optional): translation applied to the Blender object (default: (0,0,0))
            tag (str): name of the created object in Blender
        """
        super().__init__(**kwargs)
        camera_object = self.blender_camera
        camera_object.data.type = 'ORTHO'
        self.ortho_scale = ortho_scale

    @property
    def ortho_scale(self) -> float:
        return self.blender_camera.data.ortho_scale

    @ortho_scale.setter
    def ortho_scale(self, val: float):
        self.blender_camera.data.ortho_scale = val

    def distance2depth(self, distmap: np.ndarray) -> np.ndarray:
        """Convert map of camera ray lengths (distmap) to map of distances to image plane (depthmap)

        Args:
            distmap (np.ndarray): Distance map

        Returns:
            np.ndarray: Depth map
        """
        # In orthogonal camera rays are orthogonal to the image plane => distmap = depthmap
        return distmap
